#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsChannelTemplate.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 27.10.2019
# Last Modified Date: 07.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Dict, Any
from ampel.base.decorator import abstractmethod
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.ChannelModel import ChannelModel
from ampel.config.builder.FirstPassConfig import FirstPassConfig
from ampel.config.builder.AbsConfigTemplate import AbsConfigTemplate


class AbsChannelTemplate(AbsConfigTemplate, ChannelModel, abstract=True):


	def get_channel(self, logger: AmpelLogger) -> Dict[str, Any]:
		if not ChannelModel._model:
			ChannelModel._model = ChannelModel._create_model()
		return {k: v for k,v in self.__dict__.items() if k in ChannelModel._aks}

	@abstractmethod
	def get_processes(self, logger: AmpelLogger, first_pass_config: FirstPassConfig) -> List[Dict[str, Any]]:
		...

	def transfer_channel_parameters(self, process: Dict[str, Any]) -> Dict[str, Any]:
		"""
		Adds channel defined information to the provided process:
		'active, 'distrib', 'source' and 'channel'
		Note: a shallow copy of the input dict is made
		"""
		p = process.copy()

		# Inactivated channels cannot have active processes
		if not self.active:
			p['active'] = False

		# Processes inherit channel distribution/file/name infos
		p['distrib'] = self.distrib
		p['source'] = self.source
		p['channel'] = self.channel

		return p



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/builder/RemoteUnitDefinition.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.11.2019
# Last Modified Date: 15.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List
from ampel.model.StrictModel import StrictModel

class RemoteUnitDefinition(StrictModel):

	class_name: str
	base: List['str']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/abstract/AbsProcessTemplate.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 27.10.2019
# Last Modified Date: 27.10.2019
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any
from ampel.log.AmpelLogger import AmpelLogger
from ampel.base.decorator import abstractmethod
from ampel.config.builder.AbsConfigTemplate import AbsConfigTemplate


class AbsProcessTemplate(AbsConfigTemplate, abstract=True):

	@abstractmethod
	def get_process(self, config: Dict[str, Any], logger: AmpelLogger) -> Dict[str, Any]:
		...



import base64
import sys
from contextlib import contextmanager
from os import path
from pathlib import Path
from typing import Optional

import pytest
import yaml
from pytest_mock import MockerFixture

from ampel.cli.JobCommand import JobCommand
from ampel.cli.main import main
from ampel.config.AmpelConfig import AmpelConfig
from ampel.secret.AmpelVault import AmpelVault


@contextmanager
def argv_context(args: list[str]):
    argv = sys.argv
    try:
        sys.argv = args
        yield
    finally:
        sys.argv = argv


def run(args: list[str]) -> Optional[int]:
    try:
        with argv_context(args):
            main()
        return None
    except SystemExit as exit:
        return exit.code


def dump(payload, tmpdir, name) -> Path:
    f = Path(tmpdir) / name
    yaml.dump(payload, f.open("w"))
    return f


@pytest.fixture
def vault(tmpdir):
    return dump({"foo": "bar"}, tmpdir, "secrets.yml")


@pytest.fixture
def secrets():
    return {
        "str": "str",
        "key.with.bunch.of.dots": "nesty",
        "dict": {"user": "none", "password": 1.5},
        "list": [1, 2, 3.5, "flerp"],
    }


@pytest.fixture
def dir_secret_store(tmpdir, secrets):
    parent_dir = Path(tmpdir) / "secrets"
    parent_dir.mkdir()
    for k, v in secrets.items():
        dump(v, parent_dir, k)
    return parent_dir


@pytest.fixture
def schema(tmpdir):
    return dump({"name": "job", "task": [{"unit": "Nonesuch"}]}, tmpdir, "schema.yml")


def test_secrets(testing_config, vault: Path, schema: Path, mocker: MockerFixture):
    new_context_unit = mocker.patch("ampel.core.UnitLoader.UnitLoader.new_context_unit")
    assert (
        run(
            [
                "ampel",
                "job",
                "--config",
                str(testing_config),
                "--secrets",
                str(vault),
                "--schema",
                str(schema),
            ]
        )
        == None
    )
    assert new_context_unit.call_count == 1
    loader_vault: Optional[AmpelVault] = new_context_unit.call_args.kwargs[
        "context"
    ].loader.vault
    assert loader_vault is not None
    secret = loader_vault.get_named_secret("foo")
    assert secret is not None
    assert secret.get() == "bar"


def test_resources_from_env(
    testing_config,
    vault: Path,
    schema: Path,
    mocker: MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
):
    new_context_unit = mocker.patch("ampel.core.UnitLoader.UnitLoader.new_context_unit")
    monkeypatch.setenv("AMPEL_CONFIG_resource.mongo", "flerp")
    monkeypatch.setenv("AMPEL_CONFIG_resource.herp", "37")
    assert (
        run(
            [
                "ampel",
                "job",
                "--config",
                str(testing_config),
                "--secrets",
                str(vault),
                "--schema",
                str(schema),
            ]
        )
        == None
    )
    config: AmpelConfig = new_context_unit.call_args.kwargs["context"].config
    assert config.get("resource.mongo") == "flerp"
    assert config.get("resource.herp") == 37


def test_dir_secrets(dir_secret_store, secrets):
    cli_op = JobCommand()
    vault = cli_op.get_vault({"secrets": dir_secret_store})
    for k, v in secrets.items():
        assert vault.get_named_secret(k).get() == v

    assert vault.get_named_secret("nonesuch") is None, "missing secrets skipped cleanly"


def test_parameter_interpolation(
    testing_config,
    vault: Path,
    tmpdir,
):

    value = "flerpyherp"

    path = tmpdir / "token"

    schema = dump(
        {
            "name": "job",
            "parameters": [{"name": "expected_value", "value": value}],
            "task": [
                {
                    "unit": "DummyOutputUnit",
                    "config": {"value": value, "path": str(path)},
                    "outputs": {
                        "parameters": [
                            {"name": "token", "value_from": {"path": str(path)}}
                        ]
                    },
                },
                {
                    "unit": "DummyInputUnit",
                    "config": {
                        "value": "{{ task.DummyOutputUnit.outputs.parameters.token }}",
                        "expected_value": "{{ job.parameters.expected_value }}",
                    },
                },
            ],
        },
        tmpdir,
        "schema.yml",
    )

    assert (
        run(
            [
                "ampel",
                "job",
                "--config",
                str(testing_config),
                "--secrets",
                str(vault),
                "--schema",
                str(schema),
            ]
        )
        == None
    )


def test_input_artifacts(
    testing_config,
    vault: Path,
    tmpdir,
):

    value = "flerpyherb"

    path = tmpdir / "token"

    schema = dump(
        {
            "name": "job",
            "parameters": [
                {
                    "name": "expected_value",
                    "value": value,
                },
                {
                    "name": "url",
                    "value": f"https://httpbin.org/base64/{base64.b64encode(value.encode()).decode()}",
                },
            ],
            "task": [
                {
                    "unit": "DummyInputUnit",
                    "config": {
                        "value": "{{ inputs.parameters.token }}",
                        "expected_value": "{{ job.parameters.expected_value }}",
                    },
                    "inputs": {
                        "parameters": [
                            {"name": "token", "value": "{{ inputs.artifacts.token }}"}
                        ],
                        "artifacts": [
                            {
                                "name": "token",
                                "path": str(path),
                                "http": {"url": "{{ job.parameters.url }}"},
                            }
                        ],
                    },
                },
            ],
        },
        tmpdir,
        "schema.yml",
    )

    assert (
        run(
            [
                "ampel",
                "job",
                "--config",
                str(testing_config),
                "--secrets",
                str(vault),
                "--schema",
                str(schema),
            ]
        )
        == None
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t2/T2Utils.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2021
# Last Modified Date: 16.09.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from datetime import datetime
from pymongo.collection import Collection
from typing import Optional, Any, Iterable, Union, Dict, Literal, Sequence

from ampel.types import UnitId, Tag, StrictIterable, ChannelId, StockId
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.content.T2Document import T2Document
from ampel.content.JournalRecord import JournalRecord
from ampel.enum.DocumentCode import DocumentCode
from ampel.mongo.query.general import build_general_query
from ampel.util.collections import check_seq_inner_type


class T2Utils:


	def __init__(self, logger: AmpelLogger):
		self._confirm = False
		self.logger = logger


	def i_know_what_i_am_doing(self) -> "T2Utils":
		self._confirm = True
		return self


	def reset_t2s(self, col: Collection, run_id: int, soft: bool = False, cli: bool = False, **kwargs) -> int:
		"""
		:param soft: if True, 'body' will not be deleted
		"""

		if not self._confirm:
			self.logger.warn("Danger zone: please confirm you know what you are doing")
			return 0

		jrec = JournalRecord(tier=-1, run=run_id, ts=datetime.utcnow().timestamp())

		if cli:
			jrec['extra'] = {'cli': True}

		update: Dict[str, Any] = {
			"$set": {"code": DocumentCode.NEW},
			"$push": {"journal": jrec}
		}

		if not soft:
			update["$unset"] = {"body": 1}

		return col \
			.update_many(self.match_t2s(**kwargs), update) \
			.modified_count


	def get_t2s(self, col: Collection, **kwargs) -> Iterable[T2Document]:
		return col.find(
			self.match_t2s(**kwargs)
		)


	def match_t2s(self,
		unit: Optional[Union[UnitId, StrictIterable[UnitId]]] = None,
		config: Optional[Union[str, int]] = None,
		code: Optional[Union[int, Sequence[int]]] = None,
		link: Optional[Union[str, Sequence[str]]] = None,
		stock: Optional[Union[StockId, StrictIterable[StockId]]] = None,
		channel: Optional[Union[ChannelId, Dict, AllOf[ChannelId], AnyOf[ChannelId], OneOf[ChannelId]]] = None,
		tag: Optional[Dict[Literal['with', 'without'], Union[Tag, Dict, AllOf[Tag], AnyOf[Tag], OneOf[Tag]]]] = None,
		custom: Optional[Dict[str, Any]] = None,
		id_mapper: Optional[AbsIdMapper] = None,
		**kwargs
	) -> Dict[str, Any]:
		"""
		:param config: use string "null" to match null
		:param link: hex encoded bytes
		:param custom: custom match criteria, for example: {"body.result.sncosmo_info.chisqdof": {'$lte': 4}}
		"""

		if id_mapper and (isinstance(stock, str) or check_seq_inner_type(stock, str)):
			stock = id_mapper.to_ampel_id(stock) # type: ignore

		match = build_general_query(stock=stock, channel=channel, tag=tag)

		if unit:
			if isinstance(unit, (int, str)):
				match['unit'] = unit
			elif isinstance(unit, (tuple, list)):
				match['unit'] = unit[0] if len(unit) == 1 else {'$in': unit}
			else:
				raise ValueError(f"Unrecognized 'unit' argument type: {type(unit)}")

		if config:
			if isinstance(config, str) and config == "null":
				match['config'] = None
			else:
				if isinstance(config, int):
					match['config'] = config
				elif isinstance(config, (list, tuple)):
					match['config'] = {'$in': [el for el in config if config != "null"]}
					if "null" in config:
						match['config']['$in'].append(None)

		if link:
			match['link'] = bytes.fromhex(link) if isinstance(link, str) else {'$in': [bytes.fromhex(el) for el in link]}

		if custom:
			match.update(custom)

		if code is not None:
			match['code'] = code

		if kwargs.get('debug'):
			self.logger.debug("Using following matching criteria: %s" % match)

		return match



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/demo/DemoPointT2Unit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2020
# Last Modified Date: 11.09.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from typing import Union
from random import randint
from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit


class DemoPointT2Unit(AbsPointT2Unit):

	test_parameter: int = 1

	def process(self, datapoint: DataPoint) -> Union[UBson, UnitResult]:

		ret = {
			"datapoint_id": datapoint['id'],
			"time": time(),
			"test_parameter": self.test_parameter
		}

		if randint(0, 1) > 0.5:
			return ret
		else:
			return UnitResult(code=10, tag="MY_CUSTOM_TAG", body=ret)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/ContextUnit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 07.10.2019
# Last Modified Date: 13.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional
from ampel.types import Traceless, TRACELESS
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.core.AmpelContext import AmpelContext
from ampel.secret.Secret import Secret

ttf = type(Traceless)


class ContextUnit(AmpelBaseModel):
	"""
	Base class for units requiring a reference to an AmpelContext instance
	"""

	context: Traceless[AmpelContext]

	#: Private variable potentially set by UnitLoader for provenance purposes. Either:
	#: * None if provanance flag is False
	#: * 0 in case model content is not serializable
	#: * any other signed int value
	_trace_id: Optional[int] = None


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)

		d = self.__dict__
		excl = {
			k for k, v in self._annots.items()
			if type(v) is ttf and v.__metadata__[0] == TRACELESS
		}

		self._trace_content = {
			k: d[k]
			for k in sorted(d)
			if k not in excl and
			not isinstance(d[k], Secret)
		}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/template/TPLCompactT3.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.07.2021
# Last Modified Date: 20.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any, Optional, Type
from ampel.types import OneOrMany
from ampel.util.pretty import prettyjson
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.UnitModel import UnitModel
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.t3.T3ReviewUnitExecutor import T3ReviewUnitExecutor
from ampel.model.t3.T3IncludeDirective import T3IncludeDirective
from ampel.model.t3.T3DocBuilderModel import T3DocBuilderModel
from ampel.abstract.AbsProcessorTemplate import AbsProcessorTemplate


class TPLCompactT3(T3DocBuilderModel, AbsProcessorTemplate):
	"""
	Enables compact run blocks such as:

	resolve_config: True
	human_id: [unit, tag]
	execute:
	- supply: ...
	  stage: ...

	- unit: T3LuminosityDistance
	  config:
	    redshifts: [0.015, 0.035, 0.045, 0.06, 0.1]

	- unit: T3HubbleResidual
	  resolve_config: False
	  config:
	    plot_tag: ZTF


	Output:

	execute:
	- unit: T3ReviewUnitExecutor
	  config:
	    supply: ...
	    stage: ...

	- unit: T3PlainUnitExecutor
	  config:
	    resolve_config: True
	    human_id: [unit, tag]
	    target:
	      unit: T3LuminosityDistance
	      config:
	        redshifts: [0.015, 0.035, 0.045, 0.06, 0.1]

	- unit: T3PlainUnitExecutor
	  config:
	    resolve_config: False
	    human_id: [unit, tag]
	    target:
	      unit: T3HubbleResidual
	      config:
	        plot_tag: ZTF


	Note that T3DocBuilderModel parameters set at root level will be applied for each run block
	(unless dedicated overrides are defined within specific run blocks)
	"""

	include: Optional[T3IncludeDirective]
	execute: OneOrMany[dict[str, Any]]


	def _merge_confs(self, d: dict[str, Any], Klass: Type[AmpelBaseModel]) -> dict[str, Any]:

		conf = {
			k: getattr(self, k)
			for k in Klass._annots
			if hasattr(self, k) and not (k in self._defaults and getattr(self, k) == self._defaults[k])
		}

		for k in d:
			if k in Klass._annots:
				conf[k] = d[k]

		return conf



	# Mandatory override
	def get_model(self, config: dict[str, Any], logger: AmpelLogger) -> UnitModel:


		out: list[dict] = []
		units = config['unit']

		for el in [self.execute] if isinstance(self.execute, dict) else self.execute:

			if 'supply' in el and 'stage' in el:
				out.append(
					{
						'unit': "T3ReviewUnitExecutor",
						'config': self._merge_confs(el, T3ReviewUnitExecutor)
					}
				)
				continue

			if 'unit' in el:
				
				if el['unit'] not in units:
					raise ValueError(f"Unknown unit: {el['unit']}")

				if "AbsT3PlainUnit" in units[el['unit']]['base']:
					out.append(
						{
							'unit': 'T3PlainUnitExecutor',
							'config': self._merge_confs(el, T3DocBuilderModel) | {
								#'target': {'unit': el['unit'], 'config': el['config']}
								'target': {
									k: el[k] for k in el
									if k not in T3DocBuilderModel._annots
								}
							}
						}
					)
					continue

				elif "AbsT3ControlUnit" in units[el['unit']]['base']:
					out.append(el)
					continue

			raise ValueError("Run block syntax unsupported: \n" + prettyjson(el))

		conf: dict[str, Any] = {'include': self.include} if self.include else {}
		conf['execute'] = out

		return UnitModel(unit = 'T3Processor', config = conf)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/AbsStockCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2021
# Last Modified Date: 25.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional, Union, Literal
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.LoadJSONAction import LoadJSONAction
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.mongo.utils import maybe_match_array
from ampel.model.UnitModel import UnitModel
from ampel.model.time.UnixTimeModel import UnixTimeModel
from ampel.model.time.TimeStringModel import TimeStringModel
from ampel.model.time.TimeLastRunModel import TimeLastRunModel
from ampel.model.time.TimeDeltaModel import TimeDeltaModel
from ampel.model.time.TimeConstraintModel import TimeConstraintModel


class AbsStockCommand(AbsCoreCommand, abstract=True):
	"""
	Base class for commands selecting/matching stock(s)
	"""

	@staticmethod
	def get_select_args_help() -> Dict[str, str]:

		return {
			# Required
			'config': 'Path to an ampel config file (yaml/json)',
			# Optional
			'secrets': 'Path to a YAML secrets store in sops format',
			'log-profile': 'One of: default, compact, headerless, verbose, debug',
			'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
			# Selection
			'stock': 'Stock id(s) (OR matched if multi-valued)',
			'channel': 'Channel(s)',
			'created-after-ts': 'Created after unix timestamp',
			'created-after-str': 'Created after date-time iso string',
			'created-after-delta': 'Created after time delta',
			'created-after-process': 'Created after last run of process with name',
			'created-before-ts': 'Created before unix timestamp',
			'created-before-str': 'Created before date-time iso string',
			'created-before-delta': 'Created before time delta',
			'created-before-process': 'Created before last run of process with name',
			'updated-after-ts': 'Updated after unix timestamp',
			'updated-after-str': 'Updated after date-time iso string',
			'updated-after-delta': 'Updated after time delta',
			'updated-after-process': 'Updated after last run of process with name',
			'updated-before-ts': 'Updated before unix timestamp',
			'updated-before-str': 'Updated before date-time iso string',
			'updated-before-delta': 'Updated before time delta',
			'updated-before-process': 'Updated before last run of process with name',
			'custom-match': 'Custom mongodb match as JSON string (ex: {"body.aKey": {"$gt": 1}})',
		}


	def add_selection_args(self, builder: ArgParserBuilder) -> None:

		# Selection args
		builder.add_group('match', 'Stock selection arguments')
		builder.add_arg('match', "stock", action=MaybeIntAction, nargs="+")
		builder.add_x_args('match',
			{'name': 'created-before-str'}, {'name': 'created-before-ts', 'type': int},
			{'name': 'created-before-delta', 'action': LoadJSONAction},
			{'name': 'created-before-process'}
		)
		builder.add_x_args('match',
			{'name': 'created-after-str'}, {'name': 'created-after-ts', 'type': int},
			{'name': 'created-after-delta', 'action': LoadJSONAction},
			{'name': 'created-after-process'}
		)
		builder.add_x_args('match',
			{'name': 'updated-before-str'}, {'name': 'updated-before-ts', 'type': int},
			{'name': 'updated-before-delta', 'action': LoadJSONAction},
			{'name': 'updated-before-process'}
		)
		builder.add_x_args('match',
			{'name': 'updated-after-str'}, {'name': 'updated-after-ts', 'type': int},
			{'name': 'updated-after-delta', 'action': LoadJSONAction},
			{'name': 'updated-after-process'}
		)
		builder.create_logic_args('match', "channel", "Channel")
		builder.create_logic_args('match', "with-tag", "Tag")
		builder.create_logic_args('match', "without-tag", "Tag", excl=True)
		builder.add_arg('match', "custom-match", metavar="#", action=LoadJSONAction)


	def get_tag(self, args: Dict[str, Any]) -> Optional[Dict[Union[Literal['with'], Literal['without']], Dict]]:

		tag: Optional[Dict[Union[Literal['with'], Literal['without']], Dict]] = None
		if args.get('with_tag'):
			tag = {'with': args['with_tag']}
		if args.get('without_tag'):
			if tag is None:
				tag = {}
			tag['without'] = args['without_tag']
		return tag


	def build_select_model(self, args: Dict[str, Any]) -> UnitModel:

		conf = {
			"created": self.get_time_model("created", args),
			"updated": self.get_time_model("updated", args),
			'channel': args['channel'],
			'custom': args['custom_match']
		}

		if args.get('tag'):
			conf['tag'] = self.get_tag(args)

		if (stock := args.get('stock')):
			conf['custom'] = {
				'_id': stock if isinstance(stock, (int, bytes, str))
					else maybe_match_array(stock)
			}

		return UnitModel(unit="T3StockSelector", config=conf)


	def get_time_model(self, prefix: str, args: Dict[str, Any]) -> TimeConstraintModel:

		d: Dict[str, Any] = {'after': None, 'before': None}

		for when in ('after', 'before'):
			if args.get(x := f"{prefix}_{when}_ts"):
				d[when] = UnixTimeModel(match_type='unix_time', value=args[x])
			elif args.get(x := f"{prefix}_{when}_str"):
				d[when] = TimeStringModel(match_type='time_string', dateTimeStr=args[x], dateTimeFormat="%Y%m%dT%H%M%S")
			elif args.get(x := f"{prefix}_{when}_delta"):
				d[when] = TimeDeltaModel(match_type='time_delta', **args[x])
			elif args.get(x := f"{prefix}_{when}_process"):
				d[when] = TimeLastRunModel(match_type='time_last_run', process_name=args[x])

		return TimeConstraintModel(**d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/time/TimeDeltaModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 29.09.2018
# Last Modified Date: 06.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Literal
from datetime import datetime, timedelta
from ampel.model.StrictModel import StrictModel


class TimeDeltaModel(StrictModel):

	match_type: Literal['time_delta']
	days: int = 0
	seconds: int = 0
	microseconds: int = 0
	milliseconds: int = 0
	minutes: int = 0
	hours: int = 0
	weeks: int = 0

	def get_timestamp(self, **kwargs) -> float:

		dt = datetime.today() + timedelta(
			days=self.days, seconds=self.seconds, microseconds=self.microseconds,
			milliseconds=self.milliseconds, minutes=self.minutes,
			hours=self.hours, weeks=self.weeks
		)

		return dt.timestamp()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/dev/NoShaper.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 05.08.2021
# Last Modified Date: 05.08.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, List, Any, Iterable, Optional
from ampel.types import StockId
from ampel.abstract.AbsT0Unit import AbsT0Unit
from ampel.content.DataPoint import DataPoint
from ampel.log.AmpelLogger import AmpelLogger


class NoShaper(AbsT0Unit):

	# override
	logger: Optional[AmpelLogger] # type: ignore[assignment]

	# Mandatory implementation
	def process(self, arg: Iterable[Dict[str, Any]], stock: StockId) -> List[DataPoint]: # type: ignore[override]
		return arg # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/update/DBUpdatesBuffer.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 31.10.2018
# Last Modified Date: 01.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from math import inf
from multiprocessing.pool import ThreadPool
from pymongo.errors import BulkWriteError
from pymongo.collection import Collection
from pymongo import UpdateOne, InsertOne, UpdateMany
from typing import Dict, List, Any, Union, Callable, Optional, Iterable, Literal

from ampel.core.Schedulable import Schedulable
from ampel.log.utils import report_exception, report_error, convert_dollars
from ampel.log.AmpelLogger import AmpelLogger
from ampel.core.AmpelDB import AmpelDB, intcol
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry

DBOp = Union[UpdateOne, UpdateMany, InsertOne]
AmpelMainCol = Literal['stock', 't0', 't1', 't2', 't3']

# Monitoring counters
stat_db_ops = AmpelMetricsRegistry.counter(
	"ops",
	"Number of bulk operations",
	subsystem="db",
	labelnames=("col",)
)
stat_db_errors = AmpelMetricsRegistry.counter(
	"errors",
	"Number of bulk op errors",
	subsystem="db",
	labelnames=("col",)
)
stat_db_time = AmpelMetricsRegistry.histogram(
	"time",
	"Latency of bulk operations",
	unit="seconds",
	subsystem="db",
	labelnames=("col",)
)

class DBUpdatesBuffer(Schedulable):
	"""
	TODO:
	* Try to mark transient docs on error
	* Do something with self._err_db_ops

	Note1:
	Regarding multithreading: pymongo uses the standard Python socket module,
	which can drop the GIL while sending and receiving data over the network.

	Note2:
	On the advantages of buffering updates and submitting them in bulk

	-> Without buffering:
	In []:
		start=time()
		for i in range(200000):
			col.insert_one({'_id':i})
		print(f"{round(time()-start, 1)}s")
	Out []: 60.4s


	-> With buffering:
	In []:
		ops=[]
		start=time()
		for i in range(200000):
			ops.append(InsertOne({'_id': i}))
			if i % 2000 == 0:
				col.bulk_write(list(ops), ordered=False)
				ops=[]
		col.bulk_write(ops, ordered=False)
		print(f"{round(time()-start, 1)}s")
	Out []: 1.7s
	"""

	def __init__(self,
		ampel_db: AmpelDB,
		run_id: Union[int, List[int]],
		logger: AmpelLogger,
		error_callback: Optional[Callable[[], None]] = None,
		catch_signals: bool = True,
		log_doc_ids: Optional[Iterable[int]] = None,
		push_interval: Optional[float] = 3.,
		max_size: Optional[int] = None,
		threads: Optional[int] = None,
		raise_exc: bool = False
	):
		"""
		:param error_callback: callback method to be called on errors
		:param catch_signals: see Schedulable docstring
		:param log_doc_ids: logs inserted/updated document IDs for the given collections.
		Collection integer identifiers are: 't0' -> 0, 't1' -> 1, 't2' -> 2, 'stock' -> 3.
		Thus, if you want to activate this behavior for all collections, use log_doc_ids=[0, 1, 2, 3].
		Notes: 1) it this will significantly increase the log size/amount.
		2) the "_id" of T2 documents is generated server-side and bulk_write does not return it,
		so it is not avail 'as is'. We instead log the natching filter (t2id, config, link)
		which is guaranteed to be unique, so it's an equivalent.
		3) Huge values of push_interval can yield a document size overflow (16MB limit per doc)
		:param push_interval: If provided, updates will be flushed every x seconds.
		:param max_size: If provided, updates will be flushed as soon as possible after the
		number of updated in any collection (stock, t0, t1, t2) exceeds the provided limit.
		:param threads: perform the various collection bound bulk_write() operations in dedicated threads.
		The provided integer number defines the size of the thread pool. Note that no real performance gain
		was yet noticed using a meaningful value such as 4 (OSX, python 3.8.1). Since bulk_write drops the GIL,
		a multithreading-like effect already occurs without the specific use a threads.
		"""

		Schedulable.__init__(self, catch_signals=catch_signals)
		self._new_buffer()
		self.error_callback = error_callback

		self._cols: Dict[AmpelMainCol, Collection] = {
			col_name: ampel_db.get_collection(col_name)
			for col_name in self.db_ops.keys()
		}

		self.stats: Dict[AmpelMainCol, int] = {
			'stock': 0, 't0': 0, 't1': 0, 't2': 0,
		}

		self._err_db_ops: Dict[str, Any] = {k: [] for k in self.db_ops.keys()}
		self._ampel_db = ampel_db
		self.run_id = run_id
		self.logger = logger
		self.max_size = max_size
		self.log_doc_ids = set(log_doc_ids) if log_doc_ids else None

		self._autopush_asap = False
		self._block_autopush = False
		self._last_update = time()

		if push_interval:
			self.push_interval = push_interval
			self.get_scheduler() \
				.every(push_interval) \
				.seconds \
				.do(self.request_autopush)

			self._job = self.get_scheduler().jobs[0]
		else:
			self.push_interval = inf

		self.thread_pool = ThreadPool(threads) if threads else None

		self.raise_exc = raise_exc


	def _new_buffer(self) -> None:
		""" Creates new buffer for pymongo operations """

		# total number of updates (of all kinds/collections)
		self.db_ops: Dict[AmpelMainCol, List[DBOp]] = {
			't2': [], 't0': [], 'stock': [], 't1': []
		}


	def stop(self) -> None:

		super().stop()
		self.push_updates(force=True)

		if self.thread_pool:
			self.thread_pool.close()
			self.thread_pool.join()


	def add_updates(self, updates: Dict[AmpelMainCol, List[DBOp]]) -> None:
		"""
		:raises: KeyError if dict key is unknown (known keys: stock, t0, t1, t2)
		"""
		for k, v in updates.items():
			self.db_ops[k] += v


	def add_col_updates(self, col: AmpelMainCol, updates: List[DBOp]) -> None:
		""" :raises: KeyError if col is unknown (known cols: stock, t0, t1, t2) """
		self.db_ops[col] += updates


	def add_col_update(self, col: AmpelMainCol, update: DBOp) -> None:
		""" :raises: KeyError if col is unknown (known cols: stock, t0, t1, t2) """
		self.db_ops[col].append(update)


	def add_t0_update(self, update: DBOp) -> None:
		self.db_ops['t0'].append(update)


	def add_t1_update(self, update: DBOp) -> None:
		self.db_ops['t1'].append(update)


	def add_t2_update(self, update: DBOp) -> None:
		self.db_ops['t2'].append(update)


	def add_stock_update(self, update: DBOp) -> None:
		self.db_ops['stock'].append(update)


	def request_autopush(self) -> None:

		t = time()
		if t - self._last_update > self.push_interval:
			if self._block_autopush:
				self._autopush_asap = True
			else:
				self.push_updates()


	def check_push(self) -> None:
		"""
		Call this method to signal that now is a good time to push updates.
		Usually called by the AlertConsumer after the processing of an alert.
		If _autopush_asap is True, it means that self.push_interval has been reached
		and thus that updates must be pushed.
		Otherwise, if the updates buffer is capped, its size must be checked.
		If it is not yet big enough (self.max_size), then self._last_update is updated.
		By doing that, the regularly scheduled request_autopush() will be delayed as long as the AP processes alerts.
		"""

		self._block_autopush = False

		if self._autopush_asap:
			return self.push_updates()

		if self.max_size:
			for v in self.db_ops.values():
				if len(v) > self.max_size:
					return self.push_updates()


	def push_updates(self, force: bool = False) -> None:

		# Do not push updates in the middle of the processing of an alert
		if self._block_autopush:
			if not force:
				return
			self._block_autopush = False

		self._last_update = time()
		if self._autopush_asap:
			self._autopush_asap = False
			self._job._schedule_next_run()

		# Reference instance buffer locally before creating a new one
		db_ops = self.db_ops
		self._new_buffer()

		for col_name in db_ops.keys():
			if db_ops[col_name]:
				if self.thread_pool:
					self.thread_pool.starmap(
						self.call_bulk_write, ([col_name, db_ops[col_name]], )
					)
				else:
					self.call_bulk_write(col_name, db_ops[col_name])


	def call_bulk_write(self, col_name: AmpelMainCol, db_ops: List, *, extra: Optional[Dict] = None) -> None:
		"""
		:param col_name: Ampel DB collection name (ex: stock, t0, t1, t2)
		:param db_ops: list of pymongo operations
		:raises: None, but stops the AlertConsumer processing by using the method
		cancel_run() when unrecoverable exceptions occur.

		Regarding the handling of BulkWriteError:
		Concurent upserts triggers a DuplicateKeyError exception.

		https://stackoverflow.com/questions/37295648/mongoose-duplicate-key-error-with-upsert
		<quote>
			An upsert that results in a document insert is not a fully atomic operation.
			Think of the upsert as performing the following discrete steps:
				Query for the identified document to upsert.
				If the document exists, atomically update the existing document.
				Else (the document doesn't exist), atomically insert a new document
				that incorporates the query fields and the update.
		</quote>

		There are many tickets opened on the mongoDB bug tracker regarding this issue.
		One of which: https://jira.mongodb.org/browse/SERVER-14322
		where is stated:
			"It is expected that the client will take appropriate action
			upon detection of such constraint violation"

		All in all: the server behaves inappropriately, the driver won't catch those
		cases for us, so we have to do the work by ourself.

		Last: the use of SON (serialized Ocument Normalisation) is deprecated according
		to the mongoDB doc. It will be removed with pymongo 4, so we should not use it anymore.
		BUT: the offending updates (UpdateOne instances) returned by the server are
		provided as SON by BulkWriteError (array 'writeErrors' contains SON objects).
		So we have no other choice than handling with them for now.
		"""
		
		with stat_db_time.labels(col_name).time():
			try:

				# Update DB
				db_res = self._cols[col_name].bulk_write(db_ops, ordered=False)
				stat_db_ops.labels(col_name).inc(len(db_ops))

				self.logger.debug(
					None, extra=self._build_log_extra(
						col_name, db_ops, db_res.bulk_api_result, extra
					)
				)

				return

			except BulkWriteError as bwe:

				try:

					dup_key_only = True

					for err_dict in bwe.details.get('writeErrors', []):

						stat_db_errors.labels(col_name).inc()
						# 'code': 11000, 'errmsg': 'E11000 duplicate key error collection: ...
						if err_dict.get("code") == 11000:

							self.logger.info(
								f"Race condition during ingestion in '{col_name}': {err_dict}"
							)
							self.logger.flush()

							# Should no longer raise pymongo.errors.DuplicateKeyError
							self._cols[col_name].update_one(
								err_dict['op']['q'],
								err_dict['op']['u'],
								upsert=err_dict['op']['upsert']
							)
							stat_db_ops.labels(col_name).inc()

						else:

							dup_key_only = False
							self._err_db_ops[col_name].append(err_dict)

							# Try to insert doc into trouble collection (raises no exception)
							# Possible exception will be logged out to console in any case
							report_error(
								self._ampel_db, msg="BulkWriteError entry details",
								logger=self.logger, info={
									'run': self.run_id,
									'err': convert_dollars(err_dict)
								}
							)

							################################
							# TODO: better than this.
							# - Mark corresponding transients with an error flag (add channel info?)
							# - Implement something for temp DB connectivity issues ?
							################################

							if self.raise_exc:
								raise

					if dup_key_only:
						self.logger.debug(
							f"Race condition(s) recovered: {len(bwe.details.get('writeErrors'))}",
							extra=self._build_log_extra(col_name, db_ops, bwe.details, extra)
						)

						return

				except Exception as ee:
					if self.raise_exc:
						raise
					# Log exc and try to insert doc into trouble collection (raises no exception)
					report_exception(self._ampel_db, self.logger, exc=ee)

			except Exception as e:
				if self.raise_exc:
					raise
				# Log exc and try to insert doc into trouble collection (raises no exception)
				report_exception(self._ampel_db, self.logger, exc=e)

			print(f"Update of {col_name} collection has failed")
			print(db_ops)

			self._err_db_ops[col_name] += db_ops
			if self.error_callback:
				self.error_callback()


	def _build_log_extra(self,
		col_name: str, ops: List[DBOp],
		bulk_api_result: Dict[str, Any],
		extra: Optional[Dict[str, Any]] = None
	) -> Dict[str, Any]:

		ret = {
			'col': intcol[col_name],
			'ins': bulk_api_result['nInserted'],
			'ups': bulk_api_result['nUpserted'],
			'mod': bulk_api_result['nModified']
		}

		if self.log_doc_ids and ret['col'] in self.log_doc_ids:
			if ret['col'] != 2:
				ret['docs'] = [op._filter['_id'] for op in ops]
			else:
				ret['docs'] = [
					{
						'unit': op._filter['unit'],
						'config': op._filter['config'],
						'link': op._filter['link']
					}
					for op in ops
				]

		if extra:
			return {**extra, **ret}

		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/view/AbsMongoFlatMultiView.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.03.2021
# Last Modified Date: 06.10.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Any, Dict, Sequence, Union
from ampel.types import ChannelId
from ampel.base.decorator import abstractmethod
from ampel.mongo.view.AbsMongoView import AbsMongoView


class AbsMongoFlatMultiView(AbsMongoView, abstract=True):
	"""
	Handles flat AND or OR connected type of channel-based views (any number of)
	No support for nested logic schema (such as {any_of: [A, all_of: {[B, C]}]})
	hence the 'flat' in class name
	"""

	channel: Sequence[ChannelId]
	t0_has_chan: bool = True


	@abstractmethod
	def get_first_match(self) -> Dict[str, Any]:
		...


	def stock(self) -> List[Dict[str, Any]]:

		return [
			self.get_first_match(),
			{
				'$project': {
					'tag': 1,
					'name': 1,
					'body': 1,
					'channel': {'$setIntersection': [self.channel, '$channel']},
					'journal': self.morph_journal('journal'),
					'ts': {chan: f'$ts.{chan}' for chan in self.channel},
				}
			}
		]


	def t0(self) -> List[Dict[str, Any]]:

		ret = [
			{
				'$set': {
					'channel': self.get_channel_intersection(),
					'meta': self.conform_meta()
				}
			}
		]

		if self.t0_has_chan:
			ret.insert(0, self.get_first_match())

		return ret


	def t1(self) -> List[Dict[str, Any]]:

		return [
			self.get_first_match(),
			{
				'$set': {
					'channel': self.get_channel_intersection(),
					'meta': self.conform_meta()
				}
			}
		]


	def t2(self) -> List[Dict[str, Any]]:

		return [
			self.get_first_match(),
			{
				'$set': {
					'channel': self.get_channel_intersection(),
					'meta': self.conform_meta()
				}
			}
		]


	def t3(self) -> List[Dict[str, Any]]:

		return [
			self.get_first_match(),
			{
				'$set': {
					'channel': self.get_channel_intersection(),
					'meta': self.conform_meta()
				}
			}
		]


	def morph_journal(self, arg: str) -> Dict[str, Any]:
		"""
		If channel is an array, reduce its value to the intersection between
		the defined channels and the requested one (should work with AND or OR channel projections).
		"""

		or_list = [
			{'$eq': [chan, '$$d.channel']}
			for chan in self.channel
		]

		or_list.insert(0, {"$not": ["$$d.channel"]})

		return {
			'$filter': {
				'input': {
					'$map': {
						'input': f"${arg}",
						'as': "d",
						'in': {
							'$cond': {
								'if': {'$isArray': '$$d.channel'},
								'then': {
									'$let': {
										'vars': {
											'intersect': {'$setIntersection': [self.channel, '$$d.channel']},
										},
										'in': {
											'$switch': {
												'branches': [
													{
														'case': {'$eq': [{'$size': '$$intersect'}, 0]},
														'then': '$$REMOVE'
													},
													{
														'case': {'$eq': [{'$size': '$$intersect'}, 1]},
														'then': {'$mergeObjects': ['$$d', {'channel': {"$arrayElemAt": ['$$intersect', 0]}}]}
													},
													{
														'case': {'$gt': [{'$size': '$$intersect'}, 1]},
														'then': {'$mergeObjects': ['$$d', {'channel': '$$intersect'}]}
													}
												]
											}
										}
									}
								},
								'else': {
									'$cond': {
										'if': {'$or': or_list},
										'then': '$$d',
										'else': '$$REMOVE'
									},
								},
							}
						}
					}
				},
				'as': 'd',
				'cond': {'$ne': ['$$d', None]}
			}
		}


	def get_meta_cases(self, arg: str) -> List[Dict[str, Any]]:
		"""
		Note that this method is used by MongoOrView but also by MongoAndView
		because it does not make sense to require an "AND" connection for the
		channel field of meta records, the AND logical connection is used solely
		for the root 'channel' field of the ampel documents
		"""
		
		return [
			{
				'case': {
					'$or': [
						{'$eq': [f"{arg}.channel", chan]}
						for chan in self.channel
					]
				},
				'then': arg
			},
			{
				'case': {
					'$and': [
						{'$isArray': f"{arg}.channel"},
						{
							'$or': [
								{'$in': [chan, f"{arg}.channel"]}
								for chan in self.channel
							]
						}
					]
				},
				'then': {
					"$mergeObjects": [
						f"{arg}",
						{
							"channel": self.get_channel_intersection(f"{arg}.channel")
						}
					]
				}
			}
		]


	def get_channel_intersection(self, arg: str = '$channel') -> Any:

		return {
			'$let': {
				'vars': {
					'intersect': {'$setIntersection': [self.channel, arg]},
				},
				'in': {
					'$cond': {
						'if': {'$eq': [{'$size': '$$intersect'}, 1]},
						'then': {"$arrayElemAt": ['$$intersect', 0]},
						'else': '$$intersect'
					},
				}
			}
		}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/db/ShortIndexModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.04.2020
# Last Modified Date: 16.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Dict, Any
from ampel.model.StrictModel import StrictModel
from ampel.mongo.model.FieldModel import FieldModel

class ShortIndexModel(StrictModel):

	field: str
	args: Optional[Dict]

	def dict(self, **kwargs) -> Dict[str, Any]:
		if self.args:
			return {
				"index": [(self.field, 1)],
				"args": self.args
			}
		return {"index": [(self.field, 1)]}


	def get_id(self) -> str:
		"""
		Returns an indexId similar to what pymongo index_information outputs.
		Ex: [('tranId', 1), ('channel', 1)] -> tranId_1_channel_1
		"""
		return FieldModel(field=self.field).get_id()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/RunCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.03.2021
# Last Modified Date: 23.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.core.AmpelContext import AmpelContext
from argparse import ArgumentParser
from typing import List, Optional, Sequence, Dict, Any, Union
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.config.AmpelConfig import AmpelConfig
from ampel.model.ProcessModel import ProcessModel
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.log.LogFlag import LogFlag


class RunCommand(AbsCoreCommand):

	def __init__(self):
		self.parser = None

	# Mandatory implementation
	def get_parser(self, sub_op: Optional[str] = None) -> Union[ArgumentParser, AmpelArgumentParser]:

		if self.parser:
			return self.parser

		parser = AmpelArgumentParser("run")

		# Help
		parser.set_help_descr({
			"config": "Path to an ampel config file (yaml/json)",
			"process": "Process name(s) to run in the provided order (no regex)",
			"secrets": "Path to a YAML secrets store in sops format",
			"log-profile": "default, compact, headerless, verbose, debug, ..."
		})

		# Required
		parser.add_arg("config", "required", type=str)
		parser.add_arg("process", "required", nargs="+", default=None)

		# Optional
		parser.add_arg("secrets", type=str)
		parser.add_arg("log-profile", type=str, default="default")

		# Examples
		p = "ampel run --config ampel_conf.yaml "
		parser.add_example("-process process1 -log-profile verbose", prepend=p)
		parser.add_example("-process my_process -process.t0.my_process.processor.config.parameter_a 9000", prepend=p)
		parser.add_example("-process process1 process2 -db.prefix AmpelTest", prepend=p)

		self.parser = parser
		return parser


	# Mandatory implementation
	def run(self, args: Dict[str, Any], unknown_args: Sequence[str], sub_op: Optional[str] = None) -> None:

		ctx = self.get_context(args, unknown_args, ContextClass=AmpelContext)

		pms: List[ProcessModel] = [
			pm for el in args['process']
			if (pm := get_process(ctx.config, el))
		]

		if not pms:
			print("No process found with the provided name(s)")
			return

		for i, pm in enumerate(pms):

			ctx.loader \
				.new_context_unit(
					model = pm.processor,
					context = ctx,
					process_name = pm.name,
					sub_type = AbsEventUnit,
					base_log_flag = LogFlag.MANUAL_RUN,
					log_profile = args['log_profile']
				) \
				.run()

			# Print seperator
			if i < len(pms) - 1:
				print("\n"+"="*68)


# Also used by JobCommand
def get_process(config: AmpelConfig, name: str) -> Optional[ProcessModel]:
	for k in ("t0", "t1", "t2", "t3"):
		if (p := config.get(f"process.{k}.{name}", dict)):
			return ProcessModel(**p)
	return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/stage/filter/AbsT3Filter.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 12.02.2020
# Last Modified Date: 21.04.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Sequence, Iterable
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.struct.AmpelBuffer import AmpelBuffer


class AbsT3Filter(AmpelBaseModel, AmpelABC, abstract=True):

	@abstractmethod
	def filter(self, abufs: Iterable[AmpelBuffer]) -> Sequence[AmpelBuffer]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/builder/AbsConfigTemplate.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.04.2020
# Last Modified Date: 18.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional
from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelBaseModel import AmpelBaseModel


class AbsConfigTemplate(AmpelABC, AmpelBaseModel, abstract=True):
	""" Known direct subclasses: AbsProcessTemplate, AbsChannelTemplate """

	template: Optional[str]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/query/stock.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 13.01.2018
# Last Modified Date: 20.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Optional, Union, Any, Literal

from ampel.types import Tag
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.time.QueryTimeModel import QueryTimeModel
from ampel.util.logicschema import reduce_to_set
from ampel.mongo.query.general import build_general_query


def build_stock_query(
	channel: Optional[Union[int, str, Dict, AllOf, AnyOf, OneOf]] = None,
	tag: Optional[Dict[Literal['with', 'without'], Union[Tag, Dict, AllOf[Tag], AnyOf[Tag], OneOf[Tag]]]] = None,
	time_created: Optional[QueryTimeModel] = None,
	time_updated: Optional[QueryTimeModel] = None
) -> Dict[str, Any]:
	"""
	:param channel: string (one channel only) or a dict schema \
		(see :obj:`QueryMatchSchema <ampel.query.QueryMatchSchema>` \
		for syntax details). None (no criterium) means all channel are considered.

	:param with_tags: "tags" to be matched by query \
		(see :obj:`QueryMatchSchema <ampel.query.QueryMatchSchema>` syntax details). \

	:param without_tags: \
		"tags" not to be matched by query

	:param time_created: \
		match against the (channel dependant) transient document creation timestamp \
		(embedded in the transient journal)

	:param time_updated: \
		match against the (channel dependant) transient document modification timestamp \
		(embedded in the transient journal)

	:returns: query dict instance with adequate matching criteria
	:raises ValueError: build_general_query can raise ValueError \
		in case the provided dict schema structure is unsupported
	"""

	query = build_general_query(channel=channel, tag=tag)

	if time_created or time_updated:

		chans = reduce_to_set("any" if channel is None else channel)
		or_list = []

		for chan_name in chans:

			and_list = []

			if time_created:
				and_list.append({
					f'ts.{chan_name}.tied': time_created.dict()
				})

			if time_updated:
				and_list.append({
					f'ts.{chan_name}.upd': time_updated.dict()
				})

			or_list.append({'$and': and_list})

		query['$or'] = or_list

	return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/t3/T3IncludeDirective.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 17.12.2021
# Last Modified Date: 17.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pydantic import BaseModel
from typing import Optional, Union
from ampel.types import OneOrMany
from ampel.model.UnitModel import UnitModel


class T3IncludeDirective(BaseModel):
	"""
	:param session: models for AbsT3Supplier[dict] instances which populates the 'session' field of T3Store
	Examples of session information are:
	- Date and time the current process was last run
	- Number of alerts processed since then
	"""

	#: Provides Iterable[T3Document]
	docs: Optional[UnitModel]

	#: Provides session information. Unit(s) must be a subclass of AbsT3Supplier
	session: Union[None, OneOrMany[UnitModel]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/DefaultProcessController.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.04.2020
# Last Modified Date: 17.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import asyncio, datetime, logging, schedule
from functools import partial
from typing import Dict, Sequence, Any, Literal, Optional, Set

from ampel.util import concurrent
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.core.AmpelContext import AmpelContext
from ampel.abstract.AbsProcessController import AbsProcessController
from ampel.secret.AmpelVault import AmpelVault
from ampel.config.AmpelConfig import AmpelConfig
from ampel.config.ScheduleEvaluator import ScheduleEvaluator
from ampel.model.ProcessModel import ProcessModel
from ampel.core.AmpelDB import AmpelDB
from ampel.core.UnitLoader import UnitLoader

log = logging.getLogger(__name__)


class DefaultProcessController(AbsProcessController):
	"""
	Process controller based on ampel.core.Schedulable, i.e which uses
	the module 'schedule' for scheduling ampel processes. It supports:
	- running processes in a dedicated environment using multiprocessing
		if parameter 'isolate' is True in the corresponding ProcessModel
	- running processes in the current environment otherwise
	"""

	isolate: bool = False
	mp_overlap: Literal['terminate', 'wait', 'ignore', 'skip'] = 'skip'
	mp_join: int = 0


	def __init__(self, **kwargs) -> None:
		"""
		:param isolate: global isolate override (applies to all processes regardless of \
		the property 'isolate' defined in their own definition. 'isolate' is a process parameter \
		that regulates if a given process should run in its own multiprocessing environment).

		:param mp_overlap: required action if the previously scheduled process did not terminate
		before the next scheduled run is to be executed.
			* terminate: Send SIGTERM to process and join process. It allows processes
			to exit gracefully but can lead to a stall if a process does not behave nicely.
			* kill: Send SIGKILL to process and join process.
			* wait: Join the process. Beware that 'wait' will cause process scheduling to stall.
			* ignore: Run the next scheduled process nonetheless. Warning: this might exhaust resources.
			* skip: Do not run the next scheduled process.
			* skip_once: Do not run the next scheduled process, but remove the currently still active process \
			from the list of registered processes so that the next scheduled process will run for sure next time. \
			Warning: this might exhaust resources

		:param mp_join: test parameter valid only for processes with multiplier == 1, not to be used in production.
		If set to 1, scheduled mp processes will be joined.
		If set to 2, additionaly, the scheduling will be stopped if the processes returns 0/None
		"""

		super().__init__(**kwargs)
		self.scheduler = schedule.Scheduler()

		# one top-level task per ProcessModel
		# invocations of run_async_process
		self._pending_schedules: Set[asyncio.Future] = set()
		# individual process replicas
		self._processes: Dict[str, Set[asyncio.Task]] = dict()

		self._prepare_isolated_processes()


	def _prepare_isolated_processes(self) -> None:

		if self.isolate:
			for p in self.processes:
				if not p.isolate:
					p.isolate = True

		isolated_processes = [p for p in self.processes if p.isolate]

		# Note: no need to freeze config if only isolated processes are to be run since
		# each isolated process will spawn its own AmpelConfig instance in its own environment
		if isolated_processes:

			# get (serializable) ampelconfig content to be provided to mp processes
			# also: lighten the config by removing process definitions which
			# are un-necessary in this case and can be rather large
			if len(isolated_processes) == len(self.processes):

				# Having only isolated processes, we can afford to modify the config
				self.mp_config = self.config.get()
				for el in ('t0', 't1', 't2', 't3'):
					if el in self.mp_config:
						del self.mp_config['process'][el]
			else:

				# Otherwise, do a shallow copy of the first two depths
				# (and remove the 'process' key/values)
				for k, v in self.config.get().items():
					if k != 'process':
						self.mp_config[k] = {kk: vv for kk, vv in v.items()}

		# If processes are to be run in the current environment, make sure the
		# config is frozen i.e recursively cast members of the loaded config
		# to immutable types (func does nothing if config is already frozen)
		if len(isolated_processes) != len(self.processes):
			self.config.freeze()
			db = AmpelDB.new(self.config, self.vault)
			self.context = AmpelContext(
				self.config, db=db, loader=UnitLoader(self.config, db=db, vault=self.vault),
			)


	def update(self,
		config: AmpelConfig,
		vault: Optional[AmpelVault],
		processes: Sequence[ProcessModel]
	) -> None:
		self.config = config
		self.processes = processes
		self.vault = vault
		self._prepare_isolated_processes()
		self.populate_schedule(now=False)


	async def run(self) -> None:

		assert self.processes
		self.populate_schedule(now=True)
		try:
			await (task := asyncio.create_task(self.run_scheduler()))
		except asyncio.CancelledError:
			task.cancel()
			await task


	def stop(self, name: Optional[str] = None) -> None:
		"""Stop scheduling new processes."""
		self.scheduler.clear(tag=name)


	def populate_schedule(self, now: bool) -> None:
		"""
		Prepare task schedule
		
		:param now: schedule unanchored tasks (i.e. those with no at_time) now
			rather than one period from now
		"""
		self.scheduler.clear()
		evaluator = ScheduleEvaluator()
		every = lambda appointment: evaluator(self.scheduler, appointment)
		for pm in self.processes:
			for appointment in pm.schedule:
				if not appointment:
					continue
				if appointment == "super":
					from ampel.log.AmpelLogger import AmpelLogger
					AmpelLogger.get_logger().error("DefaultProcessController does not handle the schedule value 'super'")
					continue
				job = (
					every(appointment)
					.do(self.run_process, pm=pm)
					.tag(pm.name)
				)
				# Pull back the first run if the first wait time is within 10
				# seconds of the period
				if now and abs((job.next_run-datetime.datetime.now()-job.period).total_seconds()) < 10:
					job.next_run -= job.period


	def _finalize_task(self, pm: ProcessModel, future: asyncio.Future) -> None:

		try:
			result = future.result()
			if self.mp_join >= 2 and not any(bool(r) for r in result):
				for job in [job for job in self.scheduler.jobs if pm.name in job.tags]:
					self.scheduler.cancel_job(job)
		except asyncio.CancelledError:
			return
		except Exception as exc:
			AbsProcessController.process_exceptions.labels(pm.tier, pm.name).inc()
			log.warn(f"{pm.name} failed", exc_info=exc)
		finally:
			self._pending_schedules.remove(future)


	def run_process(self, pm: ProcessModel) -> None:

		if pm.isolate:
			task = asyncio.ensure_future(self.run_async_process(pm))
			self._pending_schedules.add(task)
			task.add_done_callback(partial(self._finalize_task, pm))
		else:
			self.run_sync_process(pm)


	async def run_scheduler(self) -> None:

		try:
			while self.scheduler.jobs:
				try:
					if self.mp_join >= 1 and self._pending_schedules:
						await asyncio.gather(*self._pending_schedules)
					if self.mp_join >= 2 and not self.scheduler.jobs:
						break
					await asyncio.sleep(1)
					self.scheduler.run_pending()
				except asyncio.CancelledError:
					for t in self._pending_schedules:
						t.cancel()
					raise
		finally:
			# We land here when the loop breaks due to a call to stop() or
			# this coroutine is cancelled. In either case, wait for any
			# in-flight processes to exit.
			await asyncio.gather(*self._pending_schedules)
			assert not self._pending_schedules, "all tasks removed themselves"


	def run_sync_process(self, pm: ProcessModel) -> None:
		""" Runs the provided process in the current environment """

		assert pm.multiplier == 1
		self.context.loader \
			.new_context_unit(
				model = pm.processor,
				context = self.context,
				sub_type = AbsEventUnit,
			) \
			.run()


	async def run_async_process(self, pm: ProcessModel) -> Sequence:
		"""
		Launch target process, potentially after previous invocations have
		finished
		"""
		# gather previously launched processes we might have to wait on
		if pm.name not in self._processes:
			tasks = self._processes[pm.name] = set()
		else:
			tasks = self._processes[pm.name]

		multiplier = pm.multiplier
		if self.mp_overlap in ('terminate', 'wait'):
			# wait until all previous invocations have finished
			to_wait = set(tasks)
			if self.mp_overlap == 'teminate':
				for t in to_wait:
					t.cancel()
			tasks.difference_update(to_wait)
			await asyncio.gather(*to_wait)
		elif self.mp_overlap == 'skip':
			# top up replica set
			tasks.difference_update({t for t in tasks if t.done()})
			multiplier = multiplier - len(tasks)
		elif self.mp_overlap == 'ignore':
			...
		else:
			raise ValueError(f"Unknown overlap type '{self.mp_overlap}'")

		def launch():
			counter = self.process_count.labels(pm.tier, pm.name)
			t = self.run_mp_process(
				self.mp_config,
				self.vault,
				pm.dict(),
			)
			counter.inc()
			t.add_done_callback(lambda t: counter.dec())
			return t

		to_wait = {launch() for _ in range(multiplier)}
		tasks.update(to_wait)
		try:
			return await asyncio.gather(*to_wait)
		finally:
			tasks.difference_update(to_wait)


	@staticmethod
	@concurrent.process
	def run_mp_process(
		config: Dict[str, Any],
		vault: Optional[AmpelVault],
		p: Dict[str, Any],
	) -> Any:

		pm = ProcessModel(**p)

		try:
			import setproctitle
			setproctitle.setproctitle(f"ampel.t{pm.tier}.{pm.name}")
		except Exception:
			...

		# Create new context with frozen config
		context = AmpelContext.load(
			config,
			vault=vault,
			freeze_config=True
		)

		processor = context.loader.new_context_unit(
			model = pm.processor,
			context = context,
			sub_type = AbsEventUnit,
			process_name = pm.name,
		)

		return processor.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/config/t3/ScheduleEvaluator.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: 30.09.2018
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import ast

class ScheduleEvaluator(ast.NodeVisitor):
	"""
	Safely evaluate scheduling lines of the form\n
	- `every(10).minutes`
	- `every().hour``
	- `every().day.at('10:30')`
	- `every().monday`
	- `every().wednesday.at('13:15')`
	
	Allows literal numbers, strings, calling member functions of schedule.Scheduler 
	"""
	def __call__(self, scheduler, line):
		self._scheduler = scheduler
		elem = ast.parse(line).body[0]
		return self.visit(elem)

	def generic_visit(self, node):
		raise ValueError("Illegal operation {}".format(type(node)))
	
	def visit_Constant(self, node):
		return node.n
	
	def visit_Name(self, node):
		return node.id
	
	def visit_Attribute(self, node):
		return getattr(self.visit(node.value), node.attr)
	
	def visit_Call(self, node):
		args = [self.visit(arg) for arg in node.args]
		func = self.visit(node.func)
		if isinstance(func, str):
			func = getattr(self._scheduler, func)
		return func(*args)
	
	def visit_Expr(self, node):
		return self.visit(node.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/AbsConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 03.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional, Literal
from ampel.log.AmpelLogger import AmpelLogger


class ConfigCollector(dict):

	def __init__(self,
		conf_section: str,
		content: Optional[Dict] = None,
		logger: Optional[AmpelLogger] = None,
		verbose: bool = False,
		debug: bool = False,
		tier: Optional[Literal[0, 1, 2, 3, "ops"]] = None
	) -> None:

		super().__init__(**content if content else {})
		self.verbose = verbose
		self.has_error = False
		self.conf_section = conf_section
		self.logger = AmpelLogger.get_logger() if logger is None else logger
		self.tier = f'T{tier}' if tier is not None else 'general'

		if debug:
			self.logger.info(
				f'Creating {self.__class__.__name__} collector '
				f'for T{self.tier} config section {conf_section}'
			)


	def error(self, msg: str, exc_info: Optional[Any] = None) -> None:
		self.logger.error(msg, exc_info=exc_info)
		self.has_error = True


	def missing_key(self,
		what: str, key: str, dist_name: str, register_file: Optional[str]
	) -> None:
		self.error(
			f'{what} dict is missing key "{key}" '
			f'{self.distrib_hint(dist_name, register_file)}'
		)


	def duplicated_entry(self,
		conf_key: str,
		prev_file: Optional[str] = None,
		prev_dist: Optional[str] = None,
		new_file: Optional[str] = None,
		new_dist: Optional[str] = None,
		section_detail: Optional[str] = None
	) -> None:

		from string import Template
		t = Template(
			'Duplicated $what definition: "$conf_key"\n' +
			'Previously set by $prev\n' +
			'Redefined by $new'
		)

		if prev_file is None:
			prev_file = self.get(conf_key, {}).get('source', 'unknown')

		if prev_dist is None:
			prev_dist = self.get(conf_key, {}).get('distrib', 'unknown')

		prev = self.distrib_hint(prev_dist, prev_file, False)
		new = self.distrib_hint(new_dist, new_file, False)

		self.error(
			t.substitute(
				what = section_detail if section_detail else self.conf_section,
				conf_key = conf_key,
				prev = prev,
				new = new if new else 'unknown',
			)
		)


	@staticmethod
	def distrib_hint(
		distrib: Optional[str] = None,
		file_register: Optional[str] = None,
		parenthesis: bool = True
	) -> str:
		""" Adds distribution name if available """

		ret = ''

		if file_register:
			if parenthesis:
				ret += '('
			ret += f'conf file: {file_register}'

		if distrib:
			if ret:
				return f'{ret} from distribution: {distrib}{")" if parenthesis else ""}'
			return f'(distribution: {distrib})' if parenthesis else f'distribution: {distrib}'

		if ret:
			return f'{ret})' if parenthesis else ret

		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ChannelModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.10.2019
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, List, Optional, Dict, Any
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.purge.PurgeModel import PurgeModel
#from ampel.model.ViewModel import ViewModel


class ChannelModel(AmpelBaseModel):

	channel: Union[int, str]
	version: Union[int, float, str]
	purge: PurgeModel = PurgeModel(
		content={'delay': 100, 'format': 'json', 'unify': True}, # type: ignore[arg-type]
		logs={'delay': 50, 'format': 'csv'} # type: ignore[arg-type]
	) 
	# view: str = "MongoChannelView"
	active: bool = True
	hash: Optional[int]
	distrib: Optional[str]
	source: Optional[str]
	contact: Optional[str]
	access: List[str] = []
	#: Identities allowed to access data associated with this channel
	members: Optional[List[str]]
	policy: List[str] = []




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/AmpelRegister.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.05.2020
# Last Modified Date: 04.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import bson
from time import time
from os.path import isdir, isfile
from pathlib import Path
from struct import calcsize
from typing import BinaryIO, Optional, Literal, Dict, Any, List, Union, Tuple, TypedDict

from ampel.log.AmpelLogger import AmpelLogger, VERBOSE
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.register import read_header, write_header, \
	get_inner_file_handle, get_outer_file_handle, rescale_header


class HeaderInfo(TypedDict):
	size: int
	len: int
	payload: Dict[str, Any]


class AmpelRegister(AmpelBaseModel):
	""" # noqa: E101

	General notes:
	--------------

	An ampel register file is made of three parts: signature, header and content

	1) The file signature is encoded in the first 11 bytes.
	The signature contains the header's length and size (3 bytes each)

	2) A BSON encoded dict is saved in the next x bytes, refered to as the file's header.
	The header can be updated independently of the register's content.
	The size of the header is customizable during file creation.

	3) The register's content is usually a compressed structure (zip, bzip or xz)
	which stores information in binary format. Compression can be turned off if so whished.
	(The size of a register containing 10 millions alert ids [8 bytes] and filter
	return code [1 byte] can be reduced from ~90MB to ~50MB using gzip with default settings).

	Properties:
	-----------
	- Registers can be re-opened and appended
	- Header content can be accessed or updated independently of the register's content.
	- Header updates are fast if enough space was reserved for updates in the first place.
	- Header size can be increased afterwards at the cost of having to rewrite the entire file once.
	  This happens automatically when needed.
	- Logging read and write access to the register content into the register header is supported
	- Registers can be capped (based on max content length or max number of run ids).
	  Once the limit is reached, the full register is renamed and a new one is created.

	Note: the module ampel.util.register contains tools to manually change ampel registers, such as:
	`get_header_content`, `open_file_and_write_header` and `rescale_header`

	Target file:
	------------

	:param path_base: the base folder path where to create/read register files.
	:param path_extra: an optional additional folder path (or paths) to be appended to `path_base`.
	:param full_path: when provided, all path building options will be ignored and a new
	  file handle will be opened using the provided full_path
	:param file_handle: if you provide a file handle to a register file (parameter file handle),
	  it will be used and all path options will be ignored.

	:param file_cap: registers can be capped, typically based on a maximum number of blocks.
	  When access to a full register is requested, the original file is renamed and a new register is created.
	  During files renames, an index is appended (suffix) to the original file name.
	  The current (newest) register file always comes without index suffix.
	  This index is saved into the current register header.

	  Example::
	    
	    ampel_register.bin.gz (current - newest)
	    ampel_register.bin.gz.1 (first renamed register - oldest)
	    ampel_register.bin.gz.2
	    ampel_register.bin.gz.3 (second newest)

	  Note1: file renames occur during the opening procedure of registers, no check is performed during the filling of registers
	  once a register is opened. A register can thus grow beyond the defined limit as long as a process keeps it open.


	Header:
	-------

	Limitations:

	- The maximum space reservable for the header is 2**24 bytes, i.e ~16MB.
	- integers cannot exceed 2**63 bits. Should you need to save bigger numbers,
	  please use the methods bindata_to_int, int_to_bindata from module ampel.util.bson

	:param new_header_size: either None, or an integer or a string (please read the note above).
	  
	  - None or 0: the header block size will equal the header encoded length. Choose this option if
	    the header is not meant to be updated later. Otherwise, updates will only be possible if the header size
	    does not grow (note that there is a margin allowed since header exceeding the limit are automatically
	    compressed using zlib and written to disk if the size condition is then fullfilled).
	  - integer number: the header will be allocated the specified number of bytes (for example 4096).
	  - a string: refers to a 'header margin' and must start with the character '+'.
	    This option can save space in some circumstances. The header space allocated for the header will equal
	    the length of the initial header (including all provided options such as `header_extra`) to which
	    the specified header margin will be added. For example, '+1024' means that 1024 bytes will be
	    allocated additionally to the initial header lengths for future updates. If the initial header length
	    is 100 bytes, then a header block of 1124 bytes will be created.

	:param header_extra: any extra to be included in the header under the key 'extra' (must be bson encodable)
	:param header_extra_base: any extra to be included in the header at root depth (must be bson encodable)
	:param header_update_anyway: if no update to the register is made, the default setting is that the
	  header is not updated. This settings forces header updates. For example, you might want
	  to save all run ids into the header whether or not they changed the content of the register.

	:param header_log_accesses: if True, timestamps will be recorded each time the register is opened/closed
	  along with the amount of new blocks appended to the register.
	  
	  Note 1: parameter `new_header_size` must be set when using this value.
	  Note 2: see docstring of paramter `header_update_anyway` that affects the behavior of this parameter.
	  
	  In the following example, `header_log_accesses` is responsible for creating/updating the key 'updated'::
	    
        "ts": {
            "created": 1590506868.3880599,
            "updated": [
                [1590506868.3880599, 1590509029.389, 1200],
                [1590507152.079873, 1590507295.080478, 2300]
           ]
	    }
	

	Errors:
	-------

	:raises: ValueError, FileNotFoundError
	- if neither `file_handle`, `full_path`, `path_base` exist.
	- if read access to a non-existing file is requested
	- if the target file is not a register file (existing empty files are ok)
	- if the 'struct' parameter of sub-class of this class differs with the values registered
	  in the file's header (this behavior can be deactivated from parameter `on_exist_check`)
	- during file renames (in case capped registers are used) if the target file already exists
	"""

	struct: str
	verbose: int = 0
	logger: AmpelLogger

	#: save files in <path_base>/<file>
	path_base: Optional[str]
	#: save files in <path_base>/<path_extra(s)>/<file>
	path_extra: Optional[List[str]]
	#: prefix for each file
	file_prefix: Optional[str]
	#: ignore all other path options and use this file path
	path_full: Optional[str]

	file_cap: Optional[Dict[Literal['blocks'], int]]

	#: use existing file handle
	file_handle: Optional[BinaryIO]

	#: compression scheme
	compression: Optional[Literal['gz', 'bz2', 'xz']] = 'gz'
	#: compression level
	compress_level: Optional[int]

	# General header options
	new_header_size: Optional[Union[int, str]]

	header_log_accesses: bool = False
	header_count_blocks: bool = True
	header_extra: Optional[Dict[str, Any]]
	header_extra_base: Optional[Dict[str, Any]]
	header_update_anyway: bool = False

	# New header options
	header_creation_size: Optional[int]

	# Which header key to check if file already exists
	on_exist_check: Optional[List[Union[str, Tuple[str, str]]]] = ['struct']
	on_exist_strict_check: bool = False

	def __init__(self, autoload: bool = True, **kwargs) -> None:
		""" See class docstring """

		super().__init__(**kwargs)

		if not hasattr(self, 'struct'):
			raise ValueError("Sub-classes of AmpelRegister must define static field 'struct'")

		if self.header_log_accesses and not self.new_header_size:
			raise ValueError("Parameter 'new_header_size' is required when using 'header_log_accesses'")

		if autoload:
			self.load()


	def load(self) -> None:

		if self.file_handle:
			hinfo = read_header(self.file_handle, self.logger if self.verbose > 1 else None)
			self._outer_fh = self.file_handle
		else:
			f_path = self.get_file_path()
			hinfo, self._outer_fh = get_outer_file_handle(
				f_path, write=True, logger=self.logger if self.verbose > 1 else None
			)

			if self._outer_fh is None:
				raise ValueError(f"{f_path}: cannot get file handle")

		# File exists (and is not empty)
		if hinfo:

			if self.verbose > 1:
				self.logger.debug("Header loaded", extra=hinfo['payload'])

			if self.on_exist_check:
				self.check_header(hinfo['payload'])

			if self.file_cap and self.check_rename(hinfo['payload']):
				self._outer_fh = self.rename_file(self._outer_fh, hinfo['payload'])
				hinfo = None

			else:
				self.header = hinfo

				# Update file access
				if self.header_log_accesses:
					self.register_file_access()

				# Hook for sub-classes
				self.onload_update_header()


		if hinfo is None:

			if self.verbose > 1:
				self.logger.debug("Generating new header")

			header_bytes = self.gen_new_header()
			write_header(
				self._outer_fh, header=header_bytes, hsize=self.header['size'],
				logger=self.logger if self.verbose else None
			)

		self._inner_fh = get_inner_file_handle(
			self._outer_fh, write=True,
			logger=self.logger if self.verbose > 0 else None
		)

		# Non-compressed file returns the EOF position when opened in mode 'ab'
		# compressed file return 0
		if self.compression is None:
			self._ftell = self._inner_fh.tell()

		self.header_sig = build_unsafe_dict_id(self.header['payload'])


	def check_rename(self, header: Dict[str, Any]) -> bool:
		""" override if needed """

		if not self.file_cap:
			return False

		if 'blocks' in self.file_cap:

			if header['blocks'] >= self.file_cap['blocks']:
				return True

			if self.verbose > 1:
				self.logger.debug("File rotation trigger not reached")

			return False

		self.logger.warn(f"Unknown 'file_cap' value: {self.file_cap}")

		return False


	def rename_file(self, fh: BinaryIO, header: Dict[str, Any]) -> BinaryIO:

		fh.close()
		self.file_index = header.get('findex', 0) + 1

		from os import rename
		target_file_path = f"{fh.name}.{self.file_index}"

		# we might handle this rather than raising an error in the future
		if isfile(target_file_path):
			raise ValueError(f"File rotation failure: {target_file_path} already exists")

		rename(fh.name, target_file_path)

		if self.verbose > 0:
			self.logger.info(f"Current register renamed into {target_file_path}")

		return get_outer_file_handle(
			fh.name, write=True, logger=self.logger if self.verbose > 1 else None
		)[1]


	def onload_update_header(self) -> None:
		"""
		Override if you need to update the header of an existing register.
		Ex: BaseAlertRegister adds the current run id
		"""
		pass


	def get_file_path(self) -> str:
		""" :raise: errors if sub-directories cannot be created """

		if self.path_full:
			outdir = self.path_full
		else:
			if not self.path_base:
				raise ValueError("Parameter path_base is not set")

			outdir = self.path_base
			if self.path_extra:
				outdir += '/' + '/'.join(self.path_extra)

		if not isdir(outdir):
			Path(outdir).mkdir(parents=True)

		return f"{outdir}/{self.get_file_name()}"


	def get_file_name(self) -> str:
		""" override if needed """
		return '.'.join([
			self.file_prefix or 'ampel_register',
			f'bin.{self.compression}' if self.compression else 'bin'
		])


	def check_header(self, header: Dict[str, Any]) -> None:
		"""
		:raises: ValueError is raised on mismatch between this instance value
		and the header value for the provided key
		"""

		for el in self.on_exist_check: # type: ignore[union-attr]

			if isinstance(el, tuple):
				self_key = el[0]
				hdr_key = el[1]
			else:
				self_key = hdr_key = el

			if not hasattr(self, self_key):
				raise ValueError(f"Variable {self_key} is missing, check your config")

			if hdr_key not in header:
				if self.on_exist_strict_check:
					raise ValueError(f"Attribute {hdr_key} missing in header")
				continue

			if header[hdr_key] != getattr(self, self_key):
				raise ValueError(
					f"{self.get_file_path()}: '{self_key}' mismatch: [{self.__class__.__name__}] "
					f"'{getattr(self, self_key)}' != '{header[hdr_key]}' [File header]"
				)


	def register_file_access(self,
		header: Optional[Dict[str, Any]] = None,
		use_this_time: Optional[float] = None,
		new_blocks: Optional[int] = None,
	) -> None:
		"""
		:param header: use provided header rather than self.header['payload']
		:param use_this_time: use provided time rather than time.time()
		:param new_blocks: None (default) when register is opened, 0 or an integer when register is closed
		"""
		hdr = header if header else self.header['payload']
		now = use_this_time if use_this_time else time()

		# File opened
		if new_blocks is None:
			if 'updated' in hdr['ts']:
				hdr['ts']['updated'].append([now, 0., 0])
			else: # new file
				hdr['ts']['updated'] = [[now, 0., 0]]
			return

		l = hdr['ts']['updated'][-1]
		l[1] = now
		l[2] = new_blocks

	def gen_new_header(self) -> bytes:
		"""
		Creates a new header and create instance variable self.header to reference it.
		:returns: bson encoded bytes representing the generated header
		:raises: ValueError if the generated header size exceeds user-provided bound parameters
		"""

		now = time()
		hdr: Dict[str, Any] = {
			'struct': self.struct,
			'ts': {'created': now}
		}

		if self.header_log_accesses:
			self.register_file_access(header=hdr, use_this_time=now)

		if self.header_count_blocks:
			hdr['blocks'] = 0

		if self.header_extra:
			hdr['extra'] = self.header_extra

		if self.header_extra_base:
			hdr = {**self.header_extra_base, **hdr}

		if hasattr(self, 'file_index'):
			hdr['findex'] = self.file_index
		elif self.file_cap:
			hdr['findex'] = 0

		hdr_bytes = bson.encode(hdr)
		hlen = len(hdr_bytes)

		if self.new_header_size:
			if isinstance(self.new_header_size, int):
				hsize = self.new_header_size
			elif isinstance(self.new_header_size, str):
				hsize = hlen + int(self.new_header_size[1:])
			else:
				raise ValueError("Parameter new_header_size is invalid")
		else:
			hsize = hlen

		if hlen > hsize:
			raise ValueError(
				f"Header too long ({hlen} > hsize), please check the "
				f"parameters impacting new header sizes"
			)

		self.header = HeaderInfo(size=hsize, len=hlen, payload=hdr)
		return hdr_bytes


	def __del__(self):
		""" method called when class is destroyed """
		if getattr(self, '_outer_fh', None):
			if self.verbose > 1:
				self.logger.debug("Destroying ampel register instance")
			self.close()


	def close(self, close_outer_fh: bool = True, update_header: bool = True):
		"""
		:param close_outer_fh: whether principal file handle should be closed
		:param update_header: possible overidde of default settings
		(aimed for admins working with command line)
		"""

		file_updated = 0

		# Important: zip file handle should be closed before header is updated
		if hasattr(self, '_inner_fh'):

			self._inner_fh.flush()
			if self.compression:
				file_updated = self._inner_fh.tell()
			else:
				file_updated = self._inner_fh.tell() - self._ftell

			# When no compression is used, inner_fh is the outer_fh
			if self._inner_fh != self._outer_fh:
				if self.verbose > 1:
					self.logger.debug("Closing inner file")
				self._inner_fh.close()

			self._inner_fh = None # type: ignore

		if not self._outer_fh or self._outer_fh.closed:
			self.logger.info("File handle already closed")
			return
		else:
			if self.verbose:
				self.logger.log(VERBOSE, f"Closing {self.get_file_path()}")

		if hasattr(self, 'header') and update_header:

			if file_updated or self.header_update_anyway:

				if self.header_count_blocks or self.header_log_accesses:

					new_blocks = int(file_updated / calcsize(self.struct))

					if self.header_count_blocks:
						self.header['payload']['blocks'] += new_blocks

					if self.header_log_accesses:
						self.register_file_access(new_blocks=new_blocks)

				if (
					self.header_log_accesses or self.header_count_blocks or
					build_unsafe_dict_id(self.header['payload']) != self.header_sig
				):

					if self.verbose:
						self.logger.log(VERBOSE, "Header has changed, triggering update")

					try:
						write_header(
							self._outer_fh, header=self.header['payload'], hsize=self.header['size'],
							flush=False, logger=self.logger if self.verbose else None
						)
					except ValueError:
						self.logger.warn("Header still too long, rescaling it")
						self._outer_fh.flush()
						self._outer_fh.close()
						rescale_header(
							self._outer_fh.name, new_size = self.header['size'] * 2, remove_old_file = True,
							header = self.header['payload']
						)
						self._outer_fh = None # type: ignore[assignment]

				elif self.verbose > 1:
					self.logger.debug("Header was not updated")

			elif self.verbose > 1:
				self.logger.debug("File was not updated, skipping potential header updates")

		elif self.verbose > 1:
			self.logger.debug("Header update is disabled")

		if self._outer_fh:
			self._outer_fh.flush()

		if close_outer_fh:
			if self._outer_fh:
				self._outer_fh.close()
			self._outer_fh = None # type: ignore[assignment]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/ProcessConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 09.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Union
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class ProcessConfigCollector(AbsDictConfigCollector):

	def add(self,
		arg: Dict[str, Any],
		dist_name: str,
		version: Union[str, float, int],
		register_file: str
	) -> None:

		# Doing basic validation here already
		for k in ("name", "schedule"):
			if k not in arg:
				return self.missing_key(what="Process", key=k, dist_name=dist_name, register_file=register_file)

		proc_name = arg["name"]

		if 'tier' not in arg:
			if len(self.tier) == 2 and self.tier.startswith('t') and self.tier[1].isdigit():
				arg['tier'] = int(self.tier[1])
			else:
				arg['tier'] = None

		if dist_name:
			arg['distrib'] = dist_name

		if register_file:
			arg['source'] = register_file

		if self.verbose:
			self.logger.log(VERBOSE,
				f"Adding {self.tier} process: '{proc_name}'" +
				f" from file '{register_file}'" if register_file else ""
			)

		if self.get(proc_name):
			return self.duplicated_entry(
				section_detail = f"{arg['tier']}.process",
				conf_key = proc_name, new_dist = dist_name, new_file = register_file
			)

		self.__setitem__(proc_name, arg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/test/dummy.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: 11.02.2021
# Last Modified By  : jvs

import pathlib
import time
from typing import List, Optional, Sequence, Tuple, Union, Any

from ampel.struct.UnitResult import UnitResult
from ampel.types import StockId, UBson
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsStockT2Unit import AbsStockT2Unit
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit
from ampel.abstract.AbsStateT2Unit import AbsStateT2Unit
from ampel.abstract.AbsTiedStateT2Unit import AbsTiedStateT2Unit

from ampel.content.DataPoint import DataPoint
from ampel.content.T1Document import T1Document
from ampel.view.T2DocView import T2DocView
from ampel.model.StateT2Dependency import StateT2Dependency
from ampel.abstract.AbsT0Muxer import AbsT0Muxer
from ampel.model.ingest.CompilerOptions import CompilerOptions


class Sleepy(AbsEventUnit):
    """
    A processor that does nothing (especially not touching the db, which is not
    mocked in subprocesses)
    """

    process_name: str = "ZZZzzzz"

    def run(self):
        time.sleep(1)


class DummyMuxer(AbsT0Muxer):

    #: number of points to add to provided list
    points_to_insert: int = 5

    def process(
        self, dps: List[DataPoint], stock_id: Optional[StockId] = None
    ) -> Tuple[Optional[List[DataPoint]], Optional[List[DataPoint]]]:
        """
        :returns: Tuple[datapoints to insert, datapoints to combine]
            <datapoints to insert> will be provided to a T0 ingester
            <datapoints to combine> will potentially be provided to an underlying T1 combiner
        """

        new_dps: list[DataPoint] = [
            {"id": i, "stock": stock_id or 0}
            for i in range(dps[-1]["id"] + 1, dps[-1]["id"] + 1 + self.points_to_insert)
        ]
        assert self.points_to_insert == 5
        assert len(new_dps) == self.points_to_insert
        return new_dps + dps, new_dps + dps


class DummyStockT2Unit(AbsStockT2Unit):
    def process(self, stock_doc):
        return {"id": stock_doc["stock"]}


class DummyPointT2Unit(AbsPointT2Unit):
    def process(self, datapoint):
        return {"thing": datapoint["body"]["thing"]}


class DummyStateT2Unit(AbsStateT2Unit):

    foo: int = 42

    def process(self, compound, datapoints):
        return {"len": len(datapoints)}


class DummyTiedStateT2Unit(AbsTiedStateT2Unit):

    t2_dependency = [StateT2Dependency(unit="DummyStateT2Unit")]

    def process(
        self,
        compound: T1Document,
        datapoints: Sequence[DataPoint],
        t2views: Sequence[T2DocView],
    ) -> Union[UBson, UnitResult]:
        assert t2views, "dependencies were found"
        assert len(t2views[-1].body or []) == 1
        data = t2views[-1].get_payload() or {}
        assert isinstance(data, dict)
        return {k: v * 2 for k, v in data.items()}


class DummyCompilerOptions(CompilerOptions):
    stock: dict[str, Any] = {"id_mapper": "ZTFIdMapper", "tag": "ZTF"}
    t0: dict[str, Any] = {"tag": "ZTF"}
    t1: dict[str, Any] = {"tag": "ZTF"}
    state_t2: dict[str, Any] = {"tag": "ZTF"}
    point_t2: dict[str, Any] = {"tag": "ZTF"}
    stock_t2: dict[str, Any] = {"tag": "ZTF"}


class DummyOutputUnit(AbsEventUnit):

    value: str
    path: pathlib.Path

    def run(self):
        self.path.write_text(self.value)

class DummyInputUnit(AbsEventUnit):

    value: str
    expected_value: str

    def run(self):
        assert self.value == self.expected_value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/ConfigCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 17.07.2021
# Last Modified Date: 12.10.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from argparse import ArgumentParser
from typing import Sequence, Dict, Any, Optional, Union
from ampel.log.AmpelLogger import AmpelLogger
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.config.builder.DistConfigBuilder import DistConfigBuilder


hlp = {
	'config': 'Path to an ampel config file (yaml/json)',
	# Optional
	'secrets': 'Path to a YAML secrets store in sops format',
	'out': 'Path to file where config will be saved (printed to stdout otherwise)',
	'sign': 'Append truncated file signature (last 6 digits) to filename',
	'stop-on-errors': 'by default, config building stops and raises an exception if an error occured.\n' +
		'- 2: stop on errors\n' +
		'- 1: ignore errors in first_pass_config only (will stop on morphing/scoping/template errors)\n' +
		'- 0: ignore all errors',
	'verbose': 'verbose',
	'ext-resource': 'path to resource config file (yaml) to be integrated into the final ampel config'
}


class ConfigCommand(AbsCoreCommand):
	"""
	"""

	def __init__(self):
		self.parsers = {}


	# Mandatory implementation
	def get_parser(self, sub_op: Optional[str] = None) -> Union[ArgumentParser, AmpelArgumentParser]:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = ['build', 'show']

		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'config', sub_ops, hlp, description = 'Build or show ampel config.'
			)

		builder = ArgParserBuilder("config")
		builder.add_parsers(sub_ops, hlp)

		# Required args
		builder.add_arg("build.required", "out")

		# Optional args
		builder.add_arg("optional", "secrets", default=None)
		builder.add_arg('optional', 'verbose', action="store_true")

		builder.add_arg("build.optional", "sign", action="store_true")
		builder.add_arg("build.optional", "ext-resource")
		builder.add_arg("show.optional", "pretty", action="store_true")
		builder.add_arg("build.optional", "stop-on-errors", default=2)

		# Example
		builder.add_example("build", "-out ampel_conf.yaml")
		builder.add_example("build", "-out ampel_conf.yaml -sign -verbose")
		builder.add_example("show", "-pretty -process -tier 0 -channel CHAN1")

		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]



	# Mandatory implementation
	def run(self, args: Dict[str, Any], unknown_args: Sequence[str], sub_op: Optional[str] = None) -> None:

		if sub_op == 'build':

			if not (verbose := args.get('verbose', False)):
				AmpelLogger.get_logger().info(
					"Building config, this might take a while... [use -verbose for details]"
				)

			cb = DistConfigBuilder(verbose=verbose)

			cb.load_distributions()

			cb.build_config(
				stop_on_errors = 0,
				skip_default_processes=True,
				config_validator = None,
				save = args['out'],
				ext_resource = args['ext_resource']
			)

		else:
			raise NotImplementedError("Not implemented yet")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/purge/PurgeLogsModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 18.06.2020
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Literal
from ampel.model.StrictModel import StrictModel


class PurgeLogsModel(StrictModel):
	"""
	:param delay: time of inactivity in days after which logs will be purged
	:param format: output format exported by ampel.
	- bson: (native format) use this format if you plan on working on archives with your own mongodb
	- json: the ObjectId '_id' will be replaced by the (embbed) log timestamp
	- csv: the ObjectId '_id' will be replaced by the (embbed) log timestamp
	:param compress: whether to compress the output file
	"""
	delay: int
	format: Literal['bson', 'json', 'csv']
	compress: bool = True
	header: bool = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/demo/DemoFirstPointT2Unit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2020
# Last Modified Date: 30.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from typing import Union
from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit


class DemoFirstPointT2Unit(AbsPointT2Unit):

	ingest = {'filter': 'PPSFilter', 'sort': 'jd', 'select': 'first'}

	def process(self, datapoint: DataPoint) -> Union[UBson, UnitResult]:
		return {"time": time()}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/t3/T2FilterModel.py
# License           : BSD-3-Clause
# Author            : Jakob van Santen <jakob.van.santen@desy.de>
# Date              : 02.08.2020
# Last Modified Date: 02.08.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from typing import Any, Dict
from ampel.model.StrictModel import StrictModel

class T2FilterModel(StrictModel):

	#: name of T2 unit
	unit: str
	#: Mongo match expression to apply to most recent result
	match: Dict[str, Any]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/aux/LogicOperatorFilterModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.01.2020
# Last Modified Date: 17.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import collections, operator
from pydantic import validator
from typing import Dict, Any, Callable, Type, Optional
from ampel.model.StrictModel import StrictModel

ops: Dict[str, Callable[[str, Any], bool]] = {
	'>': operator.gt,
	'<': operator.lt,
	'>=': operator.ge,
	'<=': operator.le,
	'==': operator.eq,
	'!=': operator.ne,
	'contains': operator.contains,
	'is': operator.is_,
	'is not': operator.is_not
}

class FilterCriterion(StrictModel):

	attribute: Optional[str] = None
	type: Optional[Type] = None
	operator: Callable
	value: Any

	@validator('type', pre=True)
	def load_type(cls, v):
		if isinstance(v, str):
			return getattr(collections.abc, v)
		return v

	@validator('operator', pre=True)
	def load_operator(cls, v):
		if isinstance(v, str):
			if v not in ops:
				raise ValueError(f"Unknown operator: {v}")
			return ops[v]
		return v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/db/MongoClientRoleModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 19.10.2019
# Last Modified Date: 13.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.model.StrictModel import StrictModel

class MongoClientRoleModel(StrictModel):
	r: str
	w: str



import pytest
import pathlib
import secrets

from ampel.model.job.JobModel import JobModel
from ampel.model.job.ExpressionParser import ExpressionParser


def test_resolve_parameters():
    job = JobModel(
        **{
            "name": "job",
            "task": [
                {
                    "unit": "flerp",
                    "config": {"param": "foo{{ job.parameters.param }}bar"},
                }
            ],
            "parameters": [
                {"name": "param", "value": "-biz-"},
            ],
        }
    )
    job.resolve_expressions(job.task[0].dict(), job.task[0])["config"][
        "param"
    ] == "foo-biz-bar"


def test_resolve_task_outputs(tmp_path):
    token_path = tmp_path / "token"
    token_path.write_text(token := secrets.token_hex())
    job = JobModel(
        **{
            "name": "job",
            "task": [
                {
                    "unit": "bloop",
                    "config": {"output": str(token_path)},
                    "outputs": {
                        "parameters": [
                            {"name": "token", "value_from": {"path": str(token_path)}}
                        ]
                    },
                },
                {
                    "unit": "flerp",
                    "config": {"token": "{{ task.bloop.outputs.parameters.token }}"},
                },
            ],
        }
    )
    job.resolve_expressions(job.task[1].dict(), job.task[1])["config"]["token"] == token


def test_evaluate_expression():
    with pytest.raises(ValueError):
        ExpressionParser.evaluate("job.parameters.param", {})
        ExpressionParser.evaluate(
            "thing.parameters.param", {"parameters": {"param": "foo"}}
        )
        ExpressionParser.evaluate(
            "parameters.param", {"thing": {"parameters": {"param": "foo"}}}
        )
    assert (
        ExpressionParser.evaluate(
            "job.parameters.param", {"job": {"parameters": {"param": "foo"}}}
        )
        == "foo"
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/CompilerOptions.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 08.05.2021
# Last Modified Date: 15.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any
from ampel.model.aux.AuxAliasableModel import AuxAliasableModel

empty: Dict[str, Any] = {}

class CompilerOptions(AuxAliasableModel):
	"""
	Will be merged with the options set by say IngestionHandlers (these will have priority).
	Allows for example to set default tags for given documents or to define a custom AbsIdMapper
	subclass for the stock compiler.
	"""

	t0: Dict[str, Any] = empty
	t1: Dict[str, Any] = empty
	state_t2: Dict[str, Any] = empty
	point_t2: Dict[str, Any] = empty
	stock_t2: Dict[str, Any] = empty
	stock: Dict[str, Any] = empty



from dataclasses import dataclass
from typing import List, Optional, Tuple

import psutil
from prometheus_client.metrics_core import (  # type: ignore
    CounterMetricFamily,
    GaugeMetricFamily,
    Metric,
)

from ampel.util.concurrent import _Process


@dataclass
class AmpelProcessCollector:
    """
    Collect current resource usage of subprocesses launched via
    ampel.util.concurrent
    """

    #: Name to use for main process. If None, do not collect process metrics for main process.
    name: Optional[str] = None

    def get_pids(self) -> List[Tuple[Tuple[str, str], Optional[int]]]:
        """
        Collect tuples of (labels, pid) for subprocesses, and, optionally, this process
        """
        processes: List[Tuple[Tuple[str, str], Optional[int]]] = []
        for name, replicas in _Process._active.items():
            for replica, pid in replicas.items():
                processes.append(((name, str(replica)), pid))
        if self.name:
            processes.append(((self.name, "0"), None))
        return processes

    def collect(self) -> List[Metric]:

        rss = GaugeMetricFamily(
            "ampel_resident_memory_bytes",
            "Resident memory size in bytes.",
            labels=("process", "replica"),
        )
        cpu = CounterMetricFamily(
            "ampel_cpu_seconds_total",
            "Total user and system CPU time spent in seconds.",
            labels=("process", "replica"),
        )

        try:
            for labels, pid in self.get_pids():
                try:
                    p = psutil.Process(pid)
                    with p.oneshot():
                        rss.add_metric(labels, p.memory_info().rss)
                        cpu.add_metric(
                            labels, (p.cpu_times().user + p.cpu_times().user)
                        )
                except psutil.NoSuchProcess:
                    ...
        except:
            ...
        return [rss, cpu]



from ampel.content.JournalRecord import JournalRecord
import mongomock, pytest
import pymongo
from os import environ
from pathlib import Path, PosixPath
from typing import Any
import subprocess
import json
import datetime

from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer
from ampel.dev.DevAmpelContext import DevAmpelContext

from ampel.mongo.update.MongoStockIngester import MongoStockIngester
from ampel.mongo.update.MongoT2Ingester import MongoT2Ingester
from ampel.ingest.T2Compiler import T2Compiler
from ampel.ingest.StockCompiler import StockCompiler
from ampel.log.AmpelLogger import AmpelLogger

from ampel.model.ingest.T1Combine import T1Combine
from ampel.model.ingest.T2Compute import T2Compute
from ampel.model.ingest.IngestBody import IngestBody
from ampel.model.UnitModel import UnitModel
from ampel.model.ingest.CompilerOptions import CompilerOptions
from ampel.model.ingest.IngestDirective import IngestDirective

from ampel.ingest.ChainedIngestionHandler import ChainedIngestionHandler


from ampel.test.dummy import (
    DummyStockT2Unit,
    DummyPointT2Unit,
    DummyStateT2Unit,
    DummyTiedStateT2Unit,
)


def pytest_addoption(parser):
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="run docker-based integration tests",
    )


@pytest.fixture(scope="session")
def mongod(pytestconfig):
    if "MONGO_PORT" in environ:
        yield "mongodb://localhost:{}".format(environ["MONGO_PORT"])
        return

    if not pytestconfig.getoption("--integration"):
        raise pytest.skip("integration tests require --integration flag")
    try:
        container = subprocess.check_output(
            ["docker", "run", "--rm", "-d", "-P", "mongo:4.4"]
        ).decode().strip()
    except FileNotFoundError:
        pytest.skip("integration tests require docker")
    try:
        port = json.loads(subprocess.check_output(["docker", "inspect", container]))[0][
            "NetworkSettings"
        ]["Ports"]["27017/tcp"][0]["HostPort"]
        # wait for startup
        list(pymongo.MongoClient(port=int(port)).list_databases())
        yield f"mongodb://localhost:{port}"
    finally:
        ...
        subprocess.check_call(["docker", "stop", container])


@pytest.fixture
def patch_mongo(monkeypatch):
    monkeypatch.setattr("ampel.core.AmpelDB.MongoClient", mongomock.MongoClient)
    # ignore codec_options in DataLoader
    monkeypatch.setattr("mongomock.codec_options.is_supported", lambda *args: None)


@pytest.fixture(scope="session")
def testing_config():
    return Path(__file__).parent / "test-data" / "testing-config.yaml"


@pytest.fixture(scope="session")
def core_config():
    """
    The config distributed with ampel-core
    """
    cb = DistConfigBuilder()
    for ext in "json", "yml", "yaml":
        cb.load_distrib("ampel-core", "conf", ext)
    return cb.build_config(config_validator=None)


@pytest.fixture
def mock_context(patch_mongo, testing_config: PosixPath):
    return DevAmpelContext.load(config=str(testing_config), purge_db=True)


@pytest.fixture
def integration_context(mongod, testing_config: PosixPath):
    return DevAmpelContext.load(
        config=str(testing_config),
        purge_db=True,
        custom_conf={"resource.mongo": mongod},
    )


# metafixture as suggested in https://github.com/pytest-dev/pytest/issues/349#issuecomment-189370273
@pytest.fixture(params=["mock_context", "integration_context"])
def dev_context(request):
    yield request.getfixturevalue(request.param)


@pytest.fixture
def ampel_logger():
    return AmpelLogger.get_logger()


@pytest.fixture
def ingest_stock(integration_context, ampel_logger):
    run_id = 0
    updates_buffer = DBUpdatesBuffer(integration_context.db, run_id=run_id, logger=ampel_logger)
    ingester = MongoStockIngester(
        updates_buffer=updates_buffer,
    )
    compiler = StockCompiler(
        run_id=run_id,
        tier=0,
    )
    compiler.add(
        "stockystock",
        channel="TEST_CHANNEL",
        journal=JournalRecord(tier=0, extra={"alert": 123}),
    )
    compiler.commit(ingester, datetime.datetime.now())
    ingester.updates_buffer.push_updates()
    assert integration_context.db.get_collection("stock").count_documents({}) == 1


@pytest.fixture
def ingest_stock_t2(integration_context: DevAmpelContext, ampel_logger):
    run_id = 0
    updates_buffer = DBUpdatesBuffer(integration_context.db, run_id=run_id, logger=ampel_logger)
    ingester = MongoT2Ingester(
        updates_buffer=updates_buffer,
    )
    stock_id = "stockystock"
    compiler = T2Compiler(run_id=run_id, tier=2)
    compiler.add(
        unit="DummyStockT2Unit",
        config=None,
        stock=stock_id,
        link=stock_id,
        channel="TEST_CHANNEL",
        traceid={},
    )
    compiler.commit(ingester, datetime.datetime.now().timestamp())
    ingester.updates_buffer.push_updates()
    assert integration_context.db.get_collection("t2").count_documents({}) == 1

    integration_context.register_unit(DummyStockT2Unit)


@pytest.fixture(params=["DummyStateT2Unit", "DummyPointT2Unit", "DummyStockT2Unit"])
def ingest_tied_t2(integration_context: DevAmpelContext, ampel_logger, request):
    """Create a T2 document with dependencies"""

    for unit in (
        DummyStockT2Unit,
        DummyPointT2Unit,
        DummyStateT2Unit,
        DummyTiedStateT2Unit,
    ):
        integration_context.register_unit(unit)

    dependency = request.param
    tied_config_id = integration_context.gen_config_id(
        unit="DummyTiedStateT2Unit",
        arg={"t2_dependency": [{"unit": dependency}]},
        logger=ampel_logger,
    )

    run_id = 0
    channel = "TEST_CHANNEL"
    #now = datetime.datetime.now()
    updates_buffer = DBUpdatesBuffer(integration_context.db, run_id=run_id, logger=ampel_logger)

    if "stock" in dependency.lower():
        body = IngestBody(
            stock_t2=[T2Compute(unit=dependency)],
            combine=[
                T1Combine(unit="T1SimpleCombiner", state_t2=[T2Compute(unit="DummyTiedStateT2Unit", config=tied_config_id)])  # type: ignore[arg-type]
            ],
        )
    elif "point" in dependency.lower():
        body = IngestBody(
            point_t2=[T2Compute(unit=dependency)],
            combine=[
                T1Combine(unit="T1SimpleCombiner", state_t2=[T2Compute(unit="DummyTiedStateT2Unit", config=tied_config_id)])  # type: ignore[arg-type]
            ],
        )
    else:
        body = IngestBody(
            combine=[
                T1Combine(unit="T1SimpleCombiner", state_t2=[T2Compute(unit=dependency), T2Compute(unit="DummyTiedStateT2Unit", config=tied_config_id)])  # type: ignore[arg-type]
            ]
        )

    hander = ChainedIngestionHandler(
        integration_context,
        tier=0,
        run_id=run_id,
        trace_id={},
        updates_buffer=updates_buffer,
        logger=ampel_logger,
        compiler_opts=CompilerOptions(),
        shaper=UnitModel(unit="NoShaper"),
        directives=[IngestDirective(channel=channel, ingest=body)],
    )

    datapoints: list[dict[str, Any]] = [
        {"id": i, "stock": "stockystock", "body": {"thing": i + 1}} for i in range(3)
    ]

    hander.ingest(datapoints, [(0, True)], stock_id="stockystock", jm_extra={"alert": 123})

    updates_buffer.push_updates()

    assert (
        integration_context.db.get_collection("t2").count_documents({"unit": request.param})
        == len(datapoints)
        if "point" in request.param.lower()
        else 1
    )
    assert (
        integration_context.db.get_collection("t2").count_documents(
            {"unit": "DummyTiedStateT2Unit"}
        )
        == 1
    )

    return request



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/test/test_T3Processor.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: 10.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Generator, Optional
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.struct.StockAttributes import StockAttributes
from ampel.view.SnapView import SnapView
from ampel.view.T3Store import T3Store
import pytest

from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit, T3Send
from ampel.t3.T3Processor import T3Processor


class Mutineer(AbsT3ReviewUnit):
    raise_on_process: bool = False

    def process(self, views, t3s=None):
        if self.raise_on_process:
            raise ValueError


def mutineer_process(config={}):

    return {
        "execute": [
            {
                "unit": "T3ReviewUnitExecutor",
                "config": {
                    "supply": {
                        "unit": "T3DefaultBufferSupplier",
                        "config": {
                            "select": {"unit": "T3StockSelector"},
                            "load": {
                                "unit": "T3SimpleDataLoader",
                                "config": {
                                    "directives": [{"col": "stock"}]
                                }
                            }
                        }
                    },
                    "stage": {
                        "unit": "T3SimpleStager",
                        "config": {
                            "execute": [{"unit": "Mutineer", "config": config}]
                        }
                    }
                }
            }
        ]
    }


@pytest.mark.parametrize(
    "config,expect_success",
    [
        ({}, True),
        ({"raise_on_process": True}, False),
    ]
)
def test_unit_raises_error(
    dev_context: DevAmpelContext, ingest_stock, config, expect_success
):
    """Run is marked failed if units raise an exception"""
    dev_context.register_unit(Mutineer)
    t3 = T3Processor(context=dev_context, process_name="test", raise_exc=False, **mutineer_process(config))
    t3.run()
    assert dev_context.db.get_collection("events").count_documents({}) == 1
    event = dev_context.db.get_collection("events").find_one({})
    assert event["run"] == 1
    assert event["success"] == expect_success


def test_view_generator(integration_context: DevAmpelContext, ingest_stock):

    class SendySend(AbsT3ReviewUnit):
        raise_on_process: bool = False

        def process(self, gen: Generator[SnapView, T3Send, None], t3s: Optional[T3Store] = None):
            for view in gen:
                gen.send(
                    (
                        view.id,
                        StockAttributes(
                            tag="TAGGYTAG",
                            name="floopsy",
                            journal=JournalAttributes(extra={"foo": "bar"}),
                        ),
                    )
                )

    integration_context.register_unit(SendySend)

    t3 = T3Processor(
        context=integration_context,
        raise_exc=True,
        process_name="t3",
        execute = [
            {
                "unit": "T3ReviewUnitExecutor",
                "config": {
                    "supply": {
                        "unit": "T3DefaultBufferSupplier",
                        "config": {
                            "select": {"unit": "T3StockSelector"},
                            "load": {
                                "unit": "T3SimpleDataLoader",
                                "config": {
                                    "directives": [{"col": "stock"}]
                                }
                            }
                        }
                    },
                    "stage": {
                        "unit": "T3SimpleStager",
                        "config": {
                            "execute": [{"unit": "SendySend"}]
                        }
                    }
                }
            }
        ]
    )
    t3.run()

    stock = integration_context.db.get_collection("stock").find_one()
    assert "TAGGYTAG" in stock["tag"]
    assert "floopsy" in stock["name"]
    assert len(entries := [jentry for jentry in stock["journal"] if jentry["tier"] == 3]) == 1
    jentry = entries[0]
    assert jentry["extra"] == {"foo": "bar"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/view/MongoAndView.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.03.2021
# Last Modified Date: 06.10.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any, Dict
from ampel.mongo.view.AbsMongoFlatMultiView import AbsMongoFlatMultiView

class MongoAndView(AbsMongoFlatMultiView):
	"""
	Note that althought we select ampel documents associated with *all* the underlying channels
	(based on root key 'channel') defining this view, we do not require an "AND" connection
	for the channel field of meta records (because it makes little sense)
	"""

	# Override to add other matching criteria (ex: tag-based selection)
	def get_first_match(self) -> Dict[str, Any]:
		return {'$match': {'channel': {'$all': self.channel}}}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/utils.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 24.03.2021
# Last Modified Date: 24.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.util.collections import check_seq_inner_type


def maybe_load_idmapper(args: Dict[str, Any]) -> None:
	"""
	Replaces the string id defined in args['id_mapper'] with an instance of the requested id mapper.
	Replaces potential string stock ids with their ampel ids.
	"""

	args['id_mapper'] = AuxUnitRegister.get_aux_class(
		args['id_mapper'], sub_type=AbsIdMapper
	)() if args['id_mapper'] else None

	if (
		args['id_mapper'] and args['id_mapper'] and
		(isinstance(args['stock'], str) or check_seq_inner_type(args['stock'], str))
	):
		args['stock'] = args['id_mapper'].to_ampel_id(args['stock']) # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/IngestBody.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 05.06.2020
# Last Modified Date: 27.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Sequence, Union
from ampel.model.StrictModel import StrictModel
from ampel.model.ingest.MuxModel import MuxModel
from ampel.model.ingest.T1Combine import T1Combine
from ampel.model.ingest.T1CombineCompute import T1CombineCompute
from ampel.model.ingest.T1CombineComputeNow import T1CombineComputeNow
from ampel.model.ingest.T2Compute import T2Compute


class IngestBody(StrictModel):
	"""
    stock_t2
    point_t2        <- based on input dps list
    combine         <- same
      state_t2      <- based on dps list returned by combine
      point_t2      <- same
    mux             <- based on input dps list
      insert
        point_t2    <- based on <datapoints insert> result from muxer unit (first list)
      combine
        state_t2    <- based on <datapoints combine> result from muxer unit (second list)
        point_t2    <- same
	"""

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`stocks <ampel.content.StockDocument.StockDocument>`
	stock_t2: Optional[Sequence[T2Compute]]

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`datapoints <ampel.content.DataPoint.DataPoint>`
	#: based on alert content
	point_t2: Optional[Sequence[T2Compute]]

	#: Create :class:`compounds <ampel.content.T1Document.T1Document>` from
	#: combined :class:`datapoints <ampel.content.DataPoint.DataPoint>` and the
	#: associated :class:`T2 documents <ampel.content.T2Document.T2Document>`
	combine: Optional[Sequence[Union[T1Combine, T1CombineCompute, T1CombineComputeNow]]]

	#: Include additional material (such as datapoints from the DB)
	mux: Optional[MuxModel]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/MuxModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 20.05.2021
# Last Modified Date: 27.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Sequence, Literal, Dict, Union
from ampel.model.UnitModel import UnitModel
from ampel.model.ingest.T1Combine import T1Combine
from ampel.model.ingest.T1CombineCompute import T1CombineCompute
from ampel.model.ingest.T1CombineComputeNow import T1CombineComputeNow
from ampel.model.ingest.T2Compute import T2Compute


class MuxModel(UnitModel):
	"""
	unit
	model
    insert
      point_t2    <- based on <datapoints insert> result from muxer unit (first list)
    combine
      state_t2    <- based on <datapoints combine> result from muxer unit (second list)
      point_t2    <- same
	"""

	#: Create :class:`compounds <ampel.content.T1Document.T1Document>`
	#: from :class:`datapoints <ampel.content.DataPoint.DataPoint>`
	combine: Optional[Sequence[Union[T1Combine, T1CombineCompute, T1CombineComputeNow]]]

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`datapoints <ampel.content.DataPoint.DataPoint>` based on muxer "insert" result
	insert: Optional[Dict[Literal['point_t2'], Sequence[T2Compute]]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/stage/T3DistributiveStager.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 22.04.2021
# Last Modified Date: 09.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from itertools import cycle
from typing import Optional, Generator
from multiprocessing.pool import ThreadPool

from ampel.view.T3Store import T3Store
from ampel.model.UnitModel import UnitModel
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.t3.stage.T3ThreadedStager import T3ThreadedStager


class T3DistributiveStager(T3ThreadedStager):
	"""
	Allows to execute a given unit multiple times in different parallel threads (with the same config).
	Each unit processes a subset of the initial ampel buffer stream.
	Example: 2 threads and the buffers ABCD:
	thread1 receives A, thread2 receives B, thread1 receive C, thread2 receives D
	This shall allow better performance when used in combination with T3 units that are slowed down
	by IO based operations (such as network requests to external services).
	Note that no performance gain will be obtained if the processing is CPU limited.
	"""

	#: t3 units (AbsT3ReviewUnit) to execute
	execute: UnitModel
	nthread: int = 4

	#: whether to add the thread index into log 'extra' for verbose purposes
	log_extra: bool = False


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self.t3_units = [
			self.get_unit(self.execute)
			for i in range(self.nthread)
		]


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> Optional[Generator[T3Document, None, None]]:

		try:

			ts = time()
			with ThreadPool(processes=self.nthread) as pool:

				# Create queues and generators for all instanciated t3 units
				queues, generators, async_results = self.create_threaded_generators(pool, self.t3_units, t3s)
				View = self.t3_units[0]._View
				qs = queues.values()
				iqs = cycle(qs)

				try:
					for ab in gen:
						next(iqs).put(View.of(ab, self.context.config, freeze=True))
				except RuntimeError as e:
					if "StopIteration" in str(e):
						return None
					raise e

				# Send sentinel to all threaded generators
				for q in qs:
					q.put(None) # type: ignore[arg-type]

				for i, (async_res, generator, t3_unit) in enumerate(zip(async_results, generators, self.t3_units)):

					# potential T3Record to be included in the T3Document
					if (t3_unit_result := async_res.get()):
						if (d := self.handle_t3_result(t3_unit, t3_unit_result, t3s, generator.stocks, ts)):
							if self.save_stock_ids:
								d['stock'] = generator.stocks
							yield d

					self.flush(t3_unit, extra={'thread': i} if self.log_extra else None)

		except Exception as e:
			self.flush(self.t3_units)
			self.event_hdlr.handle_error(e, self.logger)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/template/ChannelWithProcsTemplate.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 14.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.log.AmpelLogger import AmpelLogger
from typing import List, Dict, Any
from ampel.config.builder.FirstPassConfig import FirstPassConfig
from ampel.abstract.AbsChannelTemplate import AbsChannelTemplate

class ChannelWithProcsTemplate(AbsChannelTemplate):
	""" Convenience class allowing channel definitions to include processes.  """

	# Note: not using List[ProcessModel] on purpose since embedded processes
	# might need template processing as well
	process: List[Dict[str, Any]]

	def get_processes(self, logger: AmpelLogger, first_pass_config: FirstPassConfig) -> List[Dict[str, Any]]:

		# Note: not enforcing channel selection for t3 processes
		# as these could require template processing first

		return [
			self.transfer_channel_parameters(p)
			for p in self.process
		]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/t3/LoaderDirective.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.12.2019
# Last Modified Date: 02.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional, Type, Literal
from ampel.content.StockDocument import StockDocument
from ampel.content.DataPoint import DataPoint
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.model.t3.AliasableModel import AliasableModel


models = {
	"stock": StockDocument,
	"t0": DataPoint,
	"t1": T1Document,
	"t2": T2Document,
}

class LoaderDirective(AliasableModel):
	"""Specification of documents to load"""

	#: Source collection
	col: Literal["stock", "t0", "t1", "t2"]

	model: Optional[Type] # TypedDict

	#: Mongo match expression to include in the query
	query_complement: Optional[Dict[str, Any]]

	#: whether to replace init config integer hash with 'resolved' config dict
	resolve_config: bool = False

	#: whether an emtpy find() result should discard entirely the associated stock for further processing
	excluding_query: bool = False

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		if not self.model and self.col in models:
			self.model = models[self.col]
		elif self.model and not hasattr(self.model, "__annotations__"):
			raise ValueError("TypedDict expected for parameter 'model'")

	def dict(self, **kwargs):
		# do not emit model if it was equivalent to the default
		rep = super().dict(**kwargs)
		if self.model == models[self.col]:
			del rep["model"]
		return rep



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/db/MongoClientOptionsModel.py
# License           : BSD-3-Clause
# Author            : Jakob van Santen <jakob.van.santen@desy.de>
# Date              : 08.11.2020
# Last Modified Date: 08.11.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from ampel.model.StrictModel import StrictModel

class MongoClientOptionsModel(StrictModel):

	# 0 means use operating system's default socket timeout
	socketTimeoutMS: int = 0

	# 0 means use operating system's default socket timeout
	connectTimeoutMS: int = 0

	# https://github.com/mongodb/specifications/blob/master/source/server-selection/server-selection.rst#serverselectiontimeoutms
	serverSelectionTimeoutMS: int = 30000 # default is 30,000 (milliseconds)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/aux/AuxAliasableModel.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: Unspecified
# Last Modified By  : jvs

from typing import Any, Type
from ampel.model.StrictModel import StrictModel
from ampel.base.AuxUnitRegister import AuxUnitRegister


class AuxAliasableModel(StrictModel):
	"""
	A model that can be initialized from the name of an aux unit
	"""

	@classmethod
	def validate(cls: Type["AuxAliasableModel"], value: Any) -> "AuxAliasableModel":
		if isinstance(value, str):
			if value in AuxUnitRegister._defs:
				return AuxUnitRegister.get_aux_class(value).validate({})
			else:
				raise ValueError(f"{cls.__name__} '{value}' not registered")
		return super().validate(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/supply/load/T3LatestStateDataLoader.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.12.2019
# Last Modified Date: 03.04.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import collections
from bson.codec_options import CodecOptions
from typing import Iterable, Union, Iterator, Optional

from ampel.types import StockId, StrictIterable
from ampel.abstract.AbsT3Loader import AbsT3Loader
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.util.collections import to_set
from ampel.mongo.query.t1 import latest_fast_query, latest_general_query
from ampel.mongo.view.FrozenValuesDict import FrozenValuesDict


class T3LatestStateDataLoader(AbsT3Loader):
	"""
	Load only T1 and T2 documents associated with the latest state of each
	stock in the selected channels. The latest state is the compound doc with
	the largest value of ``len`` or ``added``.

	.. note::
	  
	  T3LatestStateDataLoader only loads state-bound T2 documents.

	.. seealso::
	  
	  :py:func:`ampel.db.query.t1.latest_fast_query`
	    for notes on how compounds are selected from T0
	  
	  :py:func:`ampel.db.query.t1.latest_general_query`
	    for notes on how compounds are selected from other tiers
	"""

	codec_options: Optional[CodecOptions] = CodecOptions(document_class=FrozenValuesDict)


	def __init__(self, **kwargs):

		super().__init__(**kwargs)

		self.col_t1 = self.context.db.get_collection("t1")

		for directive in self.directives:

			if directive.col == "t1":
				if directive.query_complement and '_id' in directive.query_complement:
					raise ValueError("query complement parameter '_id' cannot be used on t1 collection")

			if directive.col == "t2":
				if directive.query_complement and 'link' in directive.query_complement:
					raise ValueError("query complement parameter 'link' cannot be used on t2 collection")


	def load(self,
		stock_ids: Union[StockId, Iterator[StockId], StrictIterable[StockId]]
	) -> Iterable[AmpelBuffer]:
		"""
		Warning: if stock_ids is an Iterator, it will be fully consumed
		"""

		if isinstance(stock_ids, collections.abc.Iterator):
			stock_ids = list(stock_ids)

		# self.logger.info(f"Loading {len(stock_ids)} transients")
		states = None

		# determine latest compId of each transient
		self.logger.info("Determining latest state")

		# ids for which the fast query cannot be used (results cast into set)
		slow_ids = set(
			el['stock'] for el in self.col_t1.find(
				{
					'stock': {'$in': stock_ids},
					'tier': {'$ne': 0}
				},
				{'_id': 0, 'stock': 1}
			)
		)

		# set of transient states (see comment below for an example)
		states = set()
		set_stock_ids = to_set(stock_ids)

		# get latest state (fast mode)
		# Output example:
		# [
		# {
		#   '_id': b']\xe2H\x0f(\xbf\xca\x0b\xd3\xba\xae\x89\x0c\xb2\xd2\xae',
		#   'stock': 1810101034343026   # (ZTF18aaayyuq)
		# },
		# {
		#   '_id': b'_\xcd\xed\xa5\xe1\x16\x98\x9ai\xf6\xcb\xbd\xe7#FT',
		#   'stock': 1810101011182029   # (ZTF18aaabikt)
		# },
		# ...
		# ]
		states.update(
			[
				el['_id'] for el in self.col_t1.aggregate(
					latest_fast_query(
						list(slow_ids.symmetric_difference(set_stock_ids)),
						channel = self.channel
					)
				)
			]
		)

		# TODO: check result length ?

		# get latest state (general mode) for the remaining transients
		for slow_id in slow_ids:

			# get latest state for single transients using general query
			latest_state = next(
				self.col_t1.aggregate(
					latest_general_query(slow_id, project={'$project': {'_id': 1}})
				),
				None
			)

			# Robustness
			if latest_state is None:
				# TODO: add error flag to transient doc ?
				# TODO: add error flag to event doc
				# TODO: add doc to Ampel_troubles
				self.logger.error(
					f"Could not retrieve latest state for transient {slow_id}"
				)
				continue

			states.add(
				latest_state['_id']
			)

		# Customize T1 & T2 queries (add state query parameter)
		directives = []

		for directive in self.directives:

			if directive.col in ("t1", "t2"):

				qd = directive.copy(deep=True)
				key = 'link' if directive.col == 't2' else '_id'

				if qd.query_complement:
					qd.query_complement[key] = {"$in": list(states)}
				else:
					qd.query_complement = {key: {"$in": list(states)}}

				directives.append(qd)

			else:
				directives.append(directive)

		self.logger.info("Loading ampel data")

		# Returns AmpelBuffer instances
		return self.data_loader.load(
			stock_ids = stock_ids,
			directives = directives,
			channel = self.channel,
			codec_options = self.codec_options
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/db/IndexModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 19.10.2019
# Last Modified Date: 13.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Optional, Dict
from ampel.model.StrictModel import StrictModel
from ampel.mongo.model.FieldModel import FieldModel

class IndexModel(StrictModel):

	index: List[FieldModel]
	args: Optional[Dict]

	def get_id(self) -> str:
		"""
		Returns an indexId similar to what pymongo index_information outputs.
		Ex: [('tranId', 1), ('channel', 1)] -> tranId_1_channel_1
		"""
		return "_".join(el.get_id() for el in self.index)



import pytest
from time import time

from ampel.content.T2Document import T2Document
from ampel.core.AmpelContext import AmpelContext
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.enum.DocumentCode import DocumentCode
from contextlib import contextmanager
from pymongo.errors import OperationFailure

from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
from ampel.t2.T2Worker import T2Worker
from ampel.test.dummy import DummyPointT2Unit
from pytest_mock import MockerFixture


@contextmanager
def collect_diff(store):
    store.clear()
    before = {}
    for metric in AmpelMetricsRegistry.registry().collect():
        for sample in metric.samples:
            key = (sample.name, tuple(sample.labels.items()))
            before[key] = sample.value

    delta = {}
    yield
    for metric in AmpelMetricsRegistry.registry().collect():
        for sample in metric.samples:
            key = (sample.name, tuple(sample.labels.items()))
            delta[key] = sample.value - (before.get(key) or 0)
    store.update(delta)


def test_metrics(integration_context, ingest_stock_t2):
    t2 = T2Worker(context=integration_context, raise_exc=True, process_name="t2")

    stats = {}
    with collect_diff(stats):
        assert t2.run() == 1

    assert (
        stats[("ampel_t2_docs_processed_total", (("unit", "DummyStockT2Unit"),))] == 1
    )
    assert stats[("ampel_t2_latency_seconds_sum", (("unit", "DummyStockT2Unit"),))] > 0


def test_error_reporting(integration_context: DevAmpelContext):
    integration_context.register_unit(DummyPointT2Unit)
    # DummyPointT2Unit will raise an error on the malformed T0 doc
    integration_context.db.get_collection("t0").insert_one({"id": 42})
    channels = ["channel_a", "channel_b"]
    doc: T2Document = {
        "unit": "DummyPointT2Unit",
        "code": DocumentCode.NEW,
        "config": None,
        "col": "t0",
        "stock": 42,
        "link": 42,
        "channel": channels,
        "meta": [{"ts": 0, "tier": 2}],
    }
    integration_context.db.get_collection("t2").insert_one(doc)
    t2 = T2Worker(
        context=integration_context, raise_exc=False, process_name="t2", run_dependent_t2s=True
    )
    assert t2.run() == 1
    assert (doc := integration_context.db.get_collection("t2").find_one({}))
    assert doc["code"] == DocumentCode.EXCEPTION
    assert (trouble := integration_context.db.get_collection("troubles").find_one({}))
    assert trouble["channel"] == channels


def test_tied_t2s(integration_context, ingest_tied_t2):
    assert (num_dps := integration_context.db.get_collection("t0").count_documents({}))
    t2 = T2Worker(
        context=integration_context, raise_exc=True, process_name="t2", run_dependent_t2s=True
    )

    num_docs = t2.run()
    t2 = integration_context.db.get_collection("t2")
    assert t2.count_documents({}) == num_docs

    if "point" in ingest_tied_t2.param.lower():
        assert num_docs == 1 + num_dps
        assert (
            next(t2.find({"unit": "DummyPointT2Unit"}).sort([("link", -1)]).limit(1))[
                "body"
            ][0]["thing"]
            == 3
        )
        assert (
            next(
                t2.find({"unit": "DummyTiedStateT2Unit"}).sort([("link", -1)]).limit(1)
            )["body"][0]["thing"]
            == 2 * 3
        )
    elif "stock" in ingest_tied_t2.param.lower():
        assert num_docs == 2
        assert (
            t2.find_one({"unit": "DummyStockT2Unit"})["body"][0]["id"] == "stockystock"
        )
        assert (
            t2.find_one({"unit": "DummyTiedStateT2Unit"})["body"][0]["id"]
            == 2 * "stockystock"
        )
    else:
        assert num_docs == 2
        assert t2.find_one({"unit": "DummyStateT2Unit"})["body"][0]["len"] == num_dps
        assert (
            t2.find_one({"unit": "DummyTiedStateT2Unit"})["body"][0]["len"]
            == 2 * num_dps
        )


def test_slow_dependency(
    integration_context: AmpelContext, ingest_tied_t2, mocker: MockerFixture
):
    """
    Simulate a race conditions between parallel T2 workers
    """
    assert (num_dps := integration_context.db.get_collection("t0").count_documents({}))
    t2 = T2Worker(
        context=integration_context,
        raise_exc=True,
        process_name="t2",
        run_dependent_t2s=True,
        backoff_on_retry={"jitter": False, "factor": 10},
    )

    # num_docs = t2.run()
    col = integration_context.db.get_collection("t2")

    # set upstream docs to RUNNING, simulating the effect of
    # a parallel worker picking them up
    assert (
        col.update_many(
            {"unit": {"$not": {"$regex": ".*Tied.*"}}},
            {"$set": {"code": DocumentCode.RUNNING}},
        ).modified_count
        > 0
    )

    # should only find depdendent doc once.
    try:
        assert t2.run() == 1
    except OperationFailure as exc:
        if str(exc) == "Unrecognized expression '$last'":
            # see:
            # - https://github.com/mongomock/mongomock/pull/734
            # - https://github.com/mongomock/mongomock/pull/770
            pytest.xfail("mongomock doesn't support $last (yet)")
        else:
            raise
    assert (
        col.count_documents({"code": DocumentCode.T2_PENDING_DEPENDENCY}) == 1
    ), "exactly 1 dependent doc marked as pending"
    dependent_doc = col.find_one({"code": DocumentCode.T2_PENDING_DEPENDENCY})
    meta = dependent_doc["meta"][-1]
    assert t2.backoff_on_retry is not None
    assert meta["retry_after"] == meta["ts"] + t2.backoff_on_retry.factor
    assert t2.run() == 0, "no more docs to run"

    # run upstream docs
    while (
        doc := col.find_one_and_update(
            {"code": DocumentCode.RUNNING}, {"$set": {"code": DocumentCode.NEW}}
        )
    ) is not None:
        assert t2.run() == 1, "new doc run"
        assert (
            col.find_one({"_id": doc["_id"]})["code"] == DocumentCode.OK
        ), "new doc finished"

    assert t2.run() == 0, "no docs to run; dependent still pending"

    # blast AbsWorker one minute into the future, past the retry_after time
    ptime = mocker.patch("ampel.abstract.AbsWorker.time", return_value=time() + 60)

    assert t2.run() == 1
    assert ptime.called
    assert col.find_one({"_id": dependent_doc["_id"]})["code"] == DocumentCode.OK



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/aux/filter/PrimitiveTypeArrayFilter.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.01.2020
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Sequence, TypeVar
from ampel.model.aux.FilterCriterion import FilterCriterion
from ampel.aux.filter.AbsLogicOperatorFilter import AbsLogicOperatorFilter

T = TypeVar("T", int, str, float, str)

class PrimitiveTypeArrayFilter(AbsLogicOperatorFilter[T]):
	"""
	In []: f = PrimitiveTypeArrayFilter(filters={'operator': '>', 'value': 2})
	In []: f.apply([1, 2, 3, 4])
	Out[]: [3, 4]

	In []: f = PrimitiveTypeArrayFilter(filters={
		'all_of': [
			{'operator': '>', 'value': 2},
			{'operator': '<', 'value': 4}
		]
	})
	In []: f.apply([1, 2, 3, 4])
	Out[]: [3]
	"""

	@staticmethod
	def _apply_filter(args: Sequence[T], f: FilterCriterion) -> List[T]:
		return [s for s in args if f.operator(s, f.value)]
